import os

from .element import Element
from .function import Function

_SUCCESS, _FAIL, _ERROR, _INTERRUPT, _SKIP, _NOT_RUN = ['SUCCESS', 'FAIL', 'ERROR', 'INTERRUPT', 'SKIP', 'NOT_RUN']

class SuiteWriterTest(Function, Element):

    cls = None

    def __init__(self, suite, file):
        super(SuiteWriterTest, self).__init__(suite)
        self.file = file
        self._expected_result = _SUCCESS
        self._selected = True
        self.when_run = WhenRunHelper(self)
        self._repetitions = 1
        self.excluded_param_values = frozenset()

    def exclude_param_value(self, param_name, param_value):
        self.excluded_param_values |= {(param_name, param_value)}

    def get_full_address(self, root_path):
        filepath = os.path.join(root_path, self.file.name)
        name = self.name
        if self.cls is not None:
            name = '{.name}.{}'.format(self.cls, name) # pylint: disable=missing-format-attribute
        return '{}:{}'.format(filepath, name)

    def repeat(self, num_repetitions):
        self.add_decorator('slash.repeat({})'.format(num_repetitions))
        self.expect_repetition(num_repetitions)

    def expect_repetition(self, num_repetitions):
        self._repetitions *= num_repetitions

    def get_num_expected_repetitions(self):
        return self._repetitions

    def add_cleanup(self, **kw):
        return self.add_deferred_event('slash.add_cleanup', name='test_cleanup', **kw)

    def expect_failure(self):
        self._expect(_FAIL)

    def expect_error(self):
        self._expect(_ERROR)

    def expect_interruption(self):
        self._expect(_INTERRUPT)

    def expect_skip(self):
        self._expect(_SKIP)

    def expect_not_run(self):
        self._expect(_NOT_RUN)

    def expect_deselect(self):
        self._selected = False

    def is_selected(self):
        return self._selected

    def _expect(self, expected_result):
        if self._expected_result != _SUCCESS:
            raise NotImplementedError() # pragma: no cover
        self._expected_result = expected_result

    def get_expected_result(self):
        return self._expected_result

    def is_method_test(self):
        raise NotImplementedError() # pragma: no cover

    def is_function_test(self):
        raise NotImplementedError() # pragma: no cover

    def _write_prologue(self, code_formatter):
        self._write_event(code_formatter, 'test_start')
        super(SuiteWriterTest, self)._write_prologue(code_formatter)

    def _write_epilogue(self, code_formatter):
        self._write_event(code_formatter, 'test_end')
        super(SuiteWriterTest, self)._write_epilogue(code_formatter)

    def _get_function_name(self):
        return 'test_{}'.format(self.id)


class WhenRunHelper(object):

    def __init__(self, test):
        super(WhenRunHelper, self).__init__()
        self.test = test

    def raise_exception(self):
        self.test.append_line('raise Exception("Test exception")')
        self.test.expect_error()

    error = raise_exception

    def fail(self):
        self.test.append_line('assert False')
        self.test.expect_failure()

    def interrupt(self):
        self.test.append_line('raise KeyboardInterrupt()')
        self.test.expect_interruption()

    def skip(self, with_reason=True, decorator=False):
        if decorator:
            if with_reason:
                self.test.add_decorator('slash.skipped("reason")')
            else:
                self.test.add_decorator('slash.skipped')
        else:
            if with_reason:
                self.test.append_line('slash.skip_test("reason")')
            else:
                self.test.append_line('slash.skip_test()')
        self.test.expect_skip()



# pylint: disable=redefined-outer-name
import itertools
import os
import pytest
import shutil
from slash.plugins import manager

from slash.plugins.builtin.ci_links import Plugin

BUILD_URLS = ('https://jenkins/job/some-job/123/',
              'https://different-jenkins/another-test/9999/')
ENV_VAR_NAMES = ('MY_BUILD_URL_VARIABLE', 'ANOTHER_VARIABLE')
LOG_PATH_TEMPLATES = ('%(build_url)s/%(log_path)s',
                      'https://my-log-storage-server/%(log_path)s')
LOG_DIRS = ('logs_ci_links_plugin_test', os.path.join('some', 'other', 'directory'))


def test_build_url_not_defined(suite):
    manager.install(Plugin(), is_internal=False)
    manager.activate('ci links')
    test = suite.add_test()
    test.when_run.fail()
    summary = suite.run()
    result = summary.get_all_results_for_test(test)[0]
    log_link = result.details.all().get('log_link')
    assert log_link is None


def test_build_url_defined(suite, build_url):
    manager.install(Plugin(), is_internal=False)
    manager.activate('ci links')
    test = suite.add_test()
    test.when_run.fail()
    summary = suite.run()
    result = summary.get_all_results_for_test(test)[0]
    log_link = result.details.all().get('log_link')
    expected_log_link = '/'.join(
        (
            build_url.rstrip('/'),
            'artifact',
            *result.get_log_path().split(os.sep)
        )
    )
    assert log_link == expected_log_link


def test_adds_link_for_passing_test(suite, build_url):
    manager.install(Plugin(), is_internal=False)
    manager.activate('ci links')
    summary = suite.run()
    for test in suite:
        result = summary.get_all_results_for_test(test)[0]
        log_link = result.details.all().get('log_link')
        expected_log_link = '/'.join(
            (
                build_url.rstrip('/'),
                'artifact',
                *result.get_log_path().split(os.sep)
            )
        )
        assert log_link == expected_log_link


def test_nondefault_build_url_environment_variable(
        suite, config_override, build_url_env_var):
    env_var_name, env_var_value = build_url_env_var
    manager.install(Plugin(), is_internal=False)
    config_override(
        'plugin_config.ci_links.build_url_environment_variable',
        env_var_name
    )
    manager.activate('ci links')
    test = suite.add_test()
    test.when_run.fail()
    summary = suite.run()
    result = summary.get_all_results_for_test(test)[0]
    log_link = result.details.all().get('log_link')
    expected_log_link = '/'.join(
        (
            env_var_value.rstrip('/'),
            'artifact',
            *result.get_log_path().split(os.sep)
        )
    )
    assert log_link == expected_log_link


@pytest.mark.parametrize('template', LOG_PATH_TEMPLATES)
def test_nondefault_link_template(
        suite, config_override, build_url, template):
    manager.install(Plugin(), is_internal=False)
    config_override('plugin_config.ci_links.link_template', template)
    manager.activate('ci links')
    test = suite.add_test()
    test.when_run.fail()
    summary = suite.run()
    result = summary.get_all_results_for_test(test)[0]
    log_link = result.details.all().get('log_link')
    local_log_path = '/'.join(result.get_log_path().split(os.sep))
    expected_log_link = template % {'build_url': build_url.rstrip('/'),
                                    'log_path': local_log_path}
    assert log_link == expected_log_link


class PluginWithOverriddenBuildURLGetter(Plugin):

    build_url = 'https://overridden/build/url'

    def _get_build_url(self):
        return self.build_url


def test_overridden_build_url_getter(suite):
    plugin = PluginWithOverriddenBuildURLGetter()
    manager.install(plugin, is_internal=False)
    manager.activate('ci links')
    test = suite.add_test()
    test.when_run.fail()
    summary = suite.run()
    result = summary.get_all_results_for_test(test)[0]
    log_link = result.details.all().get('log_link')
    local_log_path = '/'.join(result.get_log_path().split(os.sep))
    expected_log_link = '%s/artifact/%s' % (plugin.build_url, local_log_path)
    assert log_link == expected_log_link


@pytest.fixture(params=BUILD_URLS)
def build_url(request):
    os.environ['BUILD_URL'] = request.param

    @request.addfinalizer
    def fin():  # pylint: disable=unused-variable
        os.environ.pop('BUILD_URL')

    return request.param


@pytest.fixture(params=itertools.product(BUILD_URLS, ENV_VAR_NAMES))
def build_url_env_var(request):
    url, env_var_name = request.param
    os.environ[env_var_name] = url

    @request.addfinalizer
    def fin():  # pylint: disable=unused-variable
        os.environ.pop(env_var_name)

    return env_var_name, url


@pytest.fixture(params=LOG_DIRS, autouse=True)
def log_dir(request, config_override):
    '''
    Create a temporary log directory relative to the working directory.
    This is used instead of the pytest built-in 'tmpdir' because
    'tmpdir' has an absolute path, and this plugin will not work with
    absolute log paths.
    '''
    os.makedirs(request.param)

    @request.addfinalizer
    def fin():  # pylint: disable=unused-variable
        shutil.rmtree(os.path.abspath(request.param))
        # shutil.rmtree does not remove the base directory when
        # removing multiple nested directories, so the base
        # directory must be removed separately
        base_dir = request.param.split(os.sep)[0]
        if os.path.exists(base_dir):
            shutil.rmtree(base_dir)

    config_override('log.root', request.param)


@pytest.fixture(scope='module', autouse=True)
def deactivate_plugin(request):

    @request.addfinalizer
    def fin():  # pylint: disable=unused-variable
        manager.deactivate('ci links')



import pytest
from slash.utils.exception_mark import ExceptionMarker, mark_exception, get_exception_mark


def test_marker(exception_class):
    exc = exception_class()
    marker = ExceptionMarker('mark')
    assert not marker.is_exception_marked(exc)
    marker.mark_exception(exc)
    assert marker.is_exception_marked(exc)


def test_marker_mark_returns_exception(exception_class):
    exc = exception_class()
    assert ExceptionMarker('mark').mark_exception(exc) is exc
    assert ExceptionMarker('mark').is_exception_marked(exc)


def test_marker_class(exception_class):
    mark_exception(exception_class, 'a', 'b')
    # Builtin types cannot be marked (on the class) either by setattr or by setting its __dict__:
    # >>> setattr(AttributeError, 'a', 'b')
    # TypeError: can't set attributes of built-in/extension type 'AttributeError'
    # >>> AttributeError.__dict__['a'] = 'b'
    # TypeError: 'mappingproxy' object does not support item assignment
    is_builtin_type = exception_class is AttributeError
    expected = None if is_builtin_type else 'b'
    assert get_exception_mark(exception_class, 'a')  == expected

    e = exception_class()
    mark_exception(e, 'c', 'd')
    assert get_exception_mark(e, 'a') == expected
    assert get_exception_mark(e, 'c') == 'd'
    assert get_exception_mark(exception_class, 'c') is None


class CustomException1(Exception):
    pass


class CustomException2(Exception):
    def __setattr__(self, name, value):
        raise Exception('Set-attr name={} value={}'.format(name, value))


@pytest.fixture(params=[CustomException1, CustomException2, AttributeError])
def exception_class(request):
    return request.param



from slash.app import Application
from slash import hooks

from slash.reporting.console_reporter import ConsoleReporter


def test_app_reporter(checkpoint):
    hooks.app_quit.register(checkpoint) # pylint: disable=no-member
    with Application() as app:
        assert isinstance(app.session.reporter, ConsoleReporter)
        assert not checkpoint.called
    assert checkpoint.called


def test_custom_reporter():

    class DummyReporter(object):
        pass

    dummy_reporter = DummyReporter()

    app = Application()
    app.set_reporter(dummy_reporter)
    with app:
        assert app.session.reporter is dummy_reporter


def test_exception_during_app_exit_debugger(checkpoint, config_override):
    @hooks.entering_debugger.register  # pylint: disable=no-member
    def callback(*_, **__):  # pylint: disable=unused-variable
        checkpoint()
        raise Exception('Entering debugger error')
    config_override('debug.enabled', True)
    with Application() as app:
        raise Exception('Some Exception')
    assert checkpoint
    assert app.exit_code != 0
    assert checkpoint.called_count == 1



import pytest


@pytest.mark.parametrize("scope", ["module", "test"])
@pytest.mark.parametrize("error_adder", [None, "add_error", "add_failure"])
def test_fixture_start_end_test(suite, suite_test, scope, error_adder):

    fixture = suite.slashconf.add_fixture(scope=scope)

    start_event = fixture.add_deferred_event(
        "this.test_start", name="fixture_test_start"
    )
    end_event = fixture.add_deferred_event("this.test_end", name="fixture_test_end")
    test_event = suite_test.add_event()

    suite_test.depend_on_fixture(fixture)

    if error_adder == "add_error":
        suite_test.when_run.error()
    elif error_adder == "add_failure":
        suite_test.when_run.fail()
    else:
        assert error_adder is None

    events = suite.run().events

    if scope != "module":
        # for module scope, the event will get run over, resulting in later `test_start` events...
        assert events[start_event].timestamp < events[test_event].timestamp
    assert events[end_event].timestamp > events[test_event].timestamp


def test_fixture_end_test_raises_exception(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():
        # pylint: disable=unused-variable
        import slash  # pylint: disable=redefined-outer-name, reimported

        @slash.fixture
        @slash.parametrize("param", [1, 2])
        def fixture(this, param):
            @this.test_end
            def test_end(*_, **__):
                1 / 0  # pylint: disable=pointless-statement

            return param

        def test_something(fixture):
            slash.context.result.data["value"] = fixture

    suite_builder.build().run().assert_all(2).exception(ZeroDivisionError).with_data(
        [{"value": 1}, {"value": 2}]
    )


def test_fixture_start_test_raises_exception(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():
        # pylint: disable=unused-variable
        import slash  # pylint: disable=redefined-outer-name, reimported

        @slash.fixture(scope='module')
        @slash.parametrize("param", [1, 2])
        def fixture(this, param):
            @this.test_start
            def test_start(*_, **__):
                1 / 0  # pylint: disable=pointless-statement

            return param

        class SomeTest(slash.Test):
            @slash.parametrize("param", [10, 20, 30])
            def test_something(self, param, fixture):  # pylint: disable=unused-argument
                slash.add_error("Not supposed to reach here").mark_fatal()

    suite_builder.build().run().assert_all(6).exception(ZeroDivisionError)


def test_fixture_start_test_raises_exception_w_before(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():
        # pylint: disable=unused-variable
        import slash  # pylint: disable=redefined-outer-name, reimported

        @slash.fixture(scope='module')
        @slash.parametrize("param", [1, 2])
        def fixture(this, param):
            @this.test_start
            def test_start(*_, **__):
                1 / 0  # pylint: disable=pointless-statement

            return param

        class SomeTest(slash.Test):
            def before(self, fixture):  # pylint: disable=unused-argument,arguments-differ
                self.data = 1

            @slash.parametrize("param", [10, 20, 30])
            def test_something(self, param):  # pylint: disable=unused-argument
                slash.add_error("Not supposed to reach here").mark_fatal()

    suite_builder.build().run().assert_all(6).exception(ZeroDivisionError)



from collections import OrderedDict
from types import TracebackType
import functools
import warnings
import inspect
import sys
import ast
import pickle
import logbook

from sentinels import NOTHING


PYPY = hasattr(sys, 'pypy_version_info')
_logger = logbook.Logger(__name__)


def check_duplicate_functions(path):
    code = None
    with open(path, 'rb') as f:
        code = f.read()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        root = ast.parse(code, filename=path)
    func_names = set()
    duplicates = set()
    for node in root.body:
        if isinstance(node, ast.FunctionDef):
            if node.name in func_names:
                duplicates.add((path, node.name, node.lineno))
            else:
                func_names.add(node.name)
    return duplicates


def wraps(func, preserve=()):
    def decorator(new_func):
        returned = functools.wraps(func)(new_func)
        returned.__wraps__ = func
        return returned
    for p in preserve:
        orig = getattr(func, p, NOTHING)
        if orig is not NOTHING:
            setattr(decorator, p, orig)
    return decorator

def unpickle(thing):
    if not thing:
        return thing
    return pickle.loads(thing.data)

def get_underlying_func(func):
    while True:
        underlying = getattr(func, "__wraps__", None)
        if underlying is None:
            return func
        func = underlying


def get_argument_names(func):
    return [arg.name for arg in get_arguments(func)]


def get_arguments_dict(func):
    returned = OrderedDict()
    for arg in get_arguments(func):
        returned[arg.name] = arg
    return returned


def get_arguments(func):
    if PYPY:
        func = get_underlying_func(func)
        spec = inspect.getargspec(func)  # pylint: disable=deprecated-method
        returned = [FunctionArgument(name=name) for name in spec.args]
    else:
        if getattr(func, '__self__', None) is not None:
            func = func.__func__ # signature() doesn't work on bound methods
        returned = [FunctionArgument.from_parameter(p) for name, p in inspect.signature(func).parameters.items()] # pylint: disable=no-member

    if returned and returned[0].name == 'self':
        returned = returned[1:]
    return returned


class FunctionArgument(object):

    def __init__(self, name, annotation=NOTHING):
        super(FunctionArgument, self).__init__()
        self.name = name
        self.annotation = annotation

    @classmethod
    def from_parameter(cls, parameter):
        annotation = parameter.annotation
        if annotation is parameter.empty:
            annotation = NOTHING
        return cls(name=parameter.name, annotation=parameter.annotation)


def call_all_raise_first(_funcs, *args, **kwargs):
    exc_info = None
    for func in _funcs:
        try:
            func(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            exc_info = sys.exc_info()
    if exc_info is not None:
        reraise(*exc_info)


def resolve_underlying_function(thing):
    """Gets the underlying (real) function for functions, wrapped functions, methods, etc.
    Returns the same object for things that are not functions
    """
    while True:
        wrapped = getattr(thing, "__func__", None) or getattr(thing, "__wrapped__", None) or getattr(thing, "__wraps__", None)
        if wrapped is None:
            break
        thing = wrapped
    return thing


def reraise(tp, value, tb=None):
    # A hacky way to check, whether we have a TracebackProxy here. Can't check directly, as
    # it would lead to circular import.
    if value.__traceback__ is not tb:
        if not isinstance(tb, TracebackType):
            tb = tb._tb # pylint: disable=protected-access
        raise value.with_traceback(tb)
    raise value


def get_underlying_classmethod_function(func):
    return func.__func__


def safe_set_attribute(obj, name, value):
    try:
        setattr(obj, name, value)
    except Exception:  # pylint: disable=broad-except
        try:
            obj.__dict__[name] = value
        except Exception:  # pylint: disable=broad-except
            _logger.debug("Failed to set attribute {!r} on {!r}", name, obj)



import os
from slash.plugins import PluginInterface, parallel_mode
from slash import context


@parallel_mode('child-only')
class Plugin(PluginInterface):
    '''
    For more information see https://slash.readthedocs.org/en/master/builtin_plugins.html#linking-to-logs-archived-by-a-ci-system
    '''

    def get_name(self):
        return 'ci links'

    def get_default_config(self):
        return {
            'build_url_environment_variable': 'BUILD_URL',
            'link_template': '%(build_url)s/artifact/%(log_path)s'
        }

    def _get_build_url(self):
        '''
        Get the URL for the current build.

        .. returns: The URL for the current build (with any trailing '/'
                    character stripped), or ``None`` if the
                    current build URL is not set.
        '''
        retval = os.environ.get(
            self.current_config.build_url_environment_variable)
        if retval is not None:
            retval = retval.rstrip('/')
        return retval

    def test_end(self):
        build_url = self._get_build_url()
        if build_url is not None:
            local_path = context.result.get_log_path()
            context.result.details.set(
                'log_link',
                self.current_config.link_template %
                {'build_url': build_url,
                 'log_path': '/'.join(local_path.split(os.sep))}
            )



from pyparsing import (
    infixNotation,
    opAssoc,
    Word,
    alphanums,
    Keyword,
    ZeroOrMore,
    Literal,
)


class Include(object):
    def __init__(self, t):
        super(Include, self).__init__()
        self.only_tags = False
        self.pattern = t[0]
        self.exact = False
        if len(t) == 2 and t[0] == "tag:":
            self.only_tags = True
            self.exact = True
            self.pattern = t[1]

    def matches(self, metadata):
        if isinstance(metadata, str):
            return self.pattern in metadata

        if metadata.tags.matches_pattern(self.pattern, exact=self.exact):
            return True
        if self.only_tags:
            return False
        return self.pattern in metadata.address

    def __repr__(self):
        return "<{}{}>".format(self.pattern, " (only tags)" if self.only_tags else "")


class BinaryMatching(object):

    aggregator = None

    def __init__(self, matchers):
        super(BinaryMatching, self).__init__()
        self.matchers = matchers

    @classmethod
    def from_tokens(cls, t):
        return cls(t[0][0::2])

    def matches(self, metadata):
        # pylint: disable=not-callable
        return self.aggregator(matcher.matches(metadata) for matcher in self.matchers)


class AndMatching(BinaryMatching):
    aggregator = all


class OrMatching(BinaryMatching):
    aggregator = any


class Exclude(object):
    def __init__(self, t):
        super(Exclude, self).__init__()
        self.matcher = t[0][1]

    def matches(self, metadata):
        return not self.matcher.matches(metadata)


word = Word(alphanums + "._,-=/:")
matcher = Literal("tag:") + ZeroOrMore(" ") + word | word
matcher.setParseAction(Include)

bool_expr = infixNotation(
    matcher,
    [
        (Keyword("not"), 1, opAssoc.RIGHT, Exclude),
        ("and", 2, opAssoc.LEFT, AndMatching.from_tokens),
        ("or", 2, opAssoc.LEFT, OrMatching.from_tokens),
    ],
)


class Matcher(object):
    def __init__(self, pattern):
        super(Matcher, self).__init__()
        self._matcher = bool_expr.parseString(pattern)[0]

    def __repr__(self):
        return repr(self._matcher)

    def matches(self, metadata):
        return self._matcher.matches(metadata)



import logbook

from . import exception_handling
from .exceptions import ExpectedExceptionNotCaught

_logger = logbook.Logger(__name__)


def _get_message(msg, description):
    if msg is None:
        return description
    return "{} ({})".format(msg, description)


class _CaughtContext(object):

    def __init__(self, message, exc_types, ensure_caught):
        if not isinstance(exc_types, tuple):
            exc_types = (exc_types, )
        self._expected_classes = exc_types
        self._caught = _CaughtException()
        self._ignore_ctx = None
        self._msg = message
        self._ensure_caught = ensure_caught

    def __enter__(self):
        self._ignore_ctx = exception_handling.thread_ignore_exception_context(self._expected_classes)
        self._ignore_ctx.__enter__()  # pylint: disable=no-member
        return self._caught

    def __exit__(self, *exc_info):
        if self._ignore_ctx:
            self._ignore_ctx.__exit__(*exc_info)  # pylint: disable=no-member
        if exc_info and exc_info != exception_handling.NO_EXC_INFO:
            e = exc_info[1]
            if isinstance(e, self._expected_classes):
                self._caught.exception = e
                return True
            return None
        msg = self._msg
        if self._msg is None:
            expected_classes = self._expected_classes
            if not isinstance(expected_classes, tuple):
                expected_classes = (expected_classes, )
            msg = "{} not raised".format("/".join(e.__name__ for e in expected_classes))
        if self._ensure_caught:
            raise ExpectedExceptionNotCaught(msg, self._expected_classes)
        _logger.debug(msg)
        return True


def assert_raises(exception_class, msg=None):
    """
    Ensures a subclass of **ARG1** leaves the wrapped context:

    >>> with assert_raises(AttributeError):
    ...     raise AttributeError()
    """
    return _CaughtContext(msg, exception_class, ensure_caught=True)


def allowing_exceptions(exception_class, msg=None):
    """
    Allow subclass of **ARG1** to be raised during context:

    >>> with allowing_exceptions(AttributeError):
    ...     raise AttributeError()
    >>> with allowing_exceptions(AttributeError):
    ...     pass
    """
    return _CaughtContext(msg, exception_class, ensure_caught=False)


def assert_almost_equal(a, b, delta=0.00000001):
    """Asserts that abs(a - b) <= delta
    """
    assert abs(a - b) <= delta


class _CaughtException(object):
    exception = None



import functools

import logbook

from ..ctx import context
from ..exceptions import SlashInternalError
from ..utils.python import call_all_raise_first

_logger = logbook.Logger(__name__)

class ScopeManager(object):

    def __init__(self, session):
        super(ScopeManager, self).__init__()
        self._session = session
        self._scopes = []
        self._last_module = self._last_test = None

    @property
    def current_scope(self):
        if not self._scopes:
            return None
        return self._scopes[-1]

    def begin_test(self, test):
        test_module = test.__slash__.module_name
        if not test_module:
            raise SlashInternalError("{!r} has no module name".format(test))

        if self._last_module is None:
            self._push_scope('session')


        if self._last_module != test_module:
            if self._last_module is not None:
                _logger.trace('Module scope has changed. Popping previous module scope')
                self._pop_scope('module')
            assert self.current_scope != 'module'
            self._push_scope('module')
        self._last_module = test_module
        self._push_scope('test')
        self._last_test = test

    def end_test(self, test):
        if test != self._last_test:
            raise SlashInternalError("Expected to pop {}, received {} instead".format(self._last_test, test))

        self._pop_scope('test')

    def get_current_stack(self):
        return self._scopes[:]

    def has_active_scopes(self):
        return bool(self._scopes)

    def _push_scope(self, scope):
        self._scopes.append(scope)
        _logger.trace('Pushed scope {0}', scope)
        self._session.fixture_store.push_scope(scope)
        self._session.cleanups.push_scope(scope)

    def _pop_scope(self, scope):
        popped = self._scopes.pop()
        _logger.trace('Popping scope {0} (expected {1})', popped, scope)
        if popped != scope:
            raise SlashInternalError('Popped scope {}, expected {}'.format(popped, scope))

        call_all_raise_first([self._session.cleanups.pop_scope, self._session.fixture_store.pop_scope],
                             scope)
        _logger.trace('Popped scope {0}', popped)

    def flush_remaining_scopes(self):
        _logger.trace('Flushing remaining scopes: {}', self._scopes)
        call_all_raise_first([functools.partial(self._pop_scope, s)
                              for s in self._scopes[::-1]])


def get_current_scope():
    if context.session is None or context.session.scope_manager is None:
        return None
    return context.session.scope_manager.current_scope



import pytest
from slash.utils.pattern_matching import Matcher
from slash.core.tagging import NO_TAGS, Tags


def test_string_patterns(suite, config_override):
    selected_test = suite[len(suite) // 2]

    config_override("run.filter_strings", [selected_test.id])

    for i, test in enumerate(suite):  # pylint: disable=unused-variable
        if test is not selected_test:
            test.expect_deselect()

    suite.run()


def test_string_patterns_and(suite, config_override):
    selected_test = suite[len(suite) // 2]
    other_test = suite[-1]

    config_override("run.filter_strings", [selected_test.id, other_test.id])

    for i, test in enumerate(suite):  # pylint: disable=unused-variable
        test.expect_deselect()

    result = suite.run(expect_session_errors=True).session.results.global_result
    assert len(result.get_errors()) == 1
    assert "no tests could be collected" in str(result.get_errors()[0]).lower()


@pytest.mark.parametrize(
    "pattern,matching,not_matching",
    [
        ("andy", ["andy", "andy_bla"], ["y", "and"]),
        ("ory", ["ory", "ory_bla"], ["y", "or"]),
        ("xandy", ["xandy", "startxandyend"], ["xy", "xxyy"]),
        ("xory", ["xory", "startxoryend"], ["xy", "xxyy", "x", "y"]),
        ("notif", ["notification", "notif"], ["kjdkjkdjkdj"]),
        ("xy", ["blapxy"], ["blap"]),
        ("not x", ["blap", "yyy"], ["blapx", "xy"]),
        ("x and y", ["xy", "yx", "x kjfldkjfd y"], ["x", "y", "xx"]),
        (
            "(x and some.string) and not z",
            ["xysome.string", "some.stringyx"],
            ["z", "xysome.stringz"],
        ),
        ("exact(x=y)(y=z)", ["exact(x=y)(y=z)"], ["exlact(x=y)(y=z)"]),
    ],
)
def test_matcher(pattern, matching, not_matching):
    matcher = Matcher(pattern)
    for string in matching:
        assert matcher.matches(
            FakeMetadata(string)
        ), "Pattern {!r} unexpectedly does not match {!r}".format(pattern, string)

    for string in not_matching:
        assert not matcher.matches(
            FakeMetadata(string)
        ), "Pattern {!r} unexpectedly matches {!r}".format(pattern, string)


def test_test_name_matcher():
    test_name = "/path/to/testfile.py:TestName"
    m = FakeMetadata(test_name)
    assert not Matcher("not {}".format(test_name)).matches(m)
    assert Matcher(test_name).matches(m)


def test_matches_tag():
    assert Matcher("bla").matches(FakeMetadata("something", {"bla": 2}))
    assert not Matcher("bla").matches(FakeMetadata("something", {"bloop": 2}))
    assert Matcher("substring").matches(
        FakeMetadata("something", {"xxxxsubstringxxx": 2})
    )


def test_matches_tag_exclusively():
    assert Matcher("tag:bla").matches(FakeMetadata("something", {"bla": 2}))
    assert not Matcher("tag:bla").matches(FakeMetadata("bla", {"bloop": 2}))


def test_matches_tag_exactly():
    matcher = Matcher("tag:bla")
    assert matcher.matches(FakeMetadata("something", {"bla": 2}))
    for non_matching_tag in ("bla1", "1bla", "1bla2"):
        assert not matcher.matches(FakeMetadata("something", {non_matching_tag: 2}))


def tet_matches_tag_whitespaces():
    assert Matcher("tag:    bla").matches(FakeMetadata("something", {"bla": 2}))
    assert not Matcher("tag: bla").matches(FakeMetadata("bla", {"bloop": 2}))


def test_matches_values():
    assert Matcher("tag:bla=2").matches(FakeMetadata("something", {"bla": 2}))
    assert not Matcher("tag:bla=2").matches(FakeMetadata("something", {"bla": 3}))
    assert Matcher("tag:bla=hello").matches(FakeMetadata("something", {"bla": "hello"}))
    assert not Matcher("tag:bla=bye").matches(
        FakeMetadata("something", {"bla": "hello"})
    )


class FakeMetadata(object):
    def __init__(self, address, tags=NO_TAGS):
        super(FakeMetadata, self).__init__()
        if isinstance(tags, dict):
            tags = Tags(tags)
        self.address = address
        self.tags = tags



from ..ctx import context
from ..exceptions import UnknownFixtures
from .fixtures.parameters import Parametrization
from . fixtures.fixture_store import nofixtures
from . import markers


def exclude(names, values):
    """
    Excludes a specific parametrization of a test from running

    :param name: can receive either a name of a parameter for this test, or a name of a fixture parameter
    :param values: must be a list of values to exclude for the given parameter
    """
    if not isinstance(values, (tuple, list)):
        raise RuntimeError('Invalid exclude values specified: must be a sequence, got {!r}'.format(values))
    if not isinstance(names, (tuple, list)):
        names = (names,)
        values = [(value,) for value in values]
    elif not isinstance(values, (tuple, list)) or any(not isinstance(item, tuple) for item in values):
        raise RuntimeError('Invalid exclude values specified for {}: {!r}'.format(', '.join(names), values))

    values = [tuple(value_set) for value_set in values]
    return markers.exclude_marker((names, values))


def is_excluded(test):
    test_func = test.get_test_function()
    if not nofixtures.is_marked(test_func):
        fixture_funcs = {fixture.fixture_func for fixture in test.get_required_fixture_objects()}
        fixture_exclusions = {fixture_func: markers.exclude_marker.get_value(fixture_func, default=None)
                              for fixture_func in fixture_funcs}
    else:
        fixture_exclusions = {}
    test_exclusions = {test_func: markers.exclude_marker.get_value(test_func, default=None)}
    for entity_func, exclusions in dict(list(fixture_exclusions.items()) + list(test_exclusions.items())).items():
        if not exclusions:
            continue
        exclusions = dict(exclusions)
        for parameter_names, value_sets in exclusions.items():
            params = []
            values = []
            for parameter_name in parameter_names:
                param = context.session.fixture_store.resolve_name(parameter_name, start_point=entity_func,
                                                                   namespace=test.get_fixture_namespace())
                if not isinstance(param, Parametrization):
                    raise UnknownFixtures('{!r} is not a parameter, and therefore cannot be the base for value exclusions'.format(parameter_name))
                params.append(param)
                try:
                    param_index = test.__slash__.variation.param_value_indices[param.info.id] #pylint: disable=no-member
                except LookupError:
                    raise UnknownFixtures('{!r} cannot be excluded for {!r}'.format(parameter_name, test))
                value = param.get_value_by_index(param_index)
                values.append(value)

            if tuple(values) in value_sets:
                return True
    return False



import pkg_resources

__version__ = pkg_resources.get_distribution('slash').version

def get_backslash_client_version():
    try:
        return pkg_resources.get_distribution('backslash').version
    except pkg_resources.DistributionNotFound:
        return None

__backslash_version__ = get_backslash_client_version()



# pylint: disable=unused-import
from .__version__ import __version__
from .cleanups import add_cleanup, add_critical_cleanup, add_success_only_cleanup, get_current_cleanup_phase, is_in_cleanup
from .conf import config
from .ctx import context
from .ctx import g, session, test
from .core.scope_manager import get_current_scope
from .core.session import Session
from .core.tagging import tag
# assertions
from . import assertions
should = assertions
from .assertions import allowing_exceptions, assert_almost_equal, assert_raises
from .core.test import Test
from .core.test import abstract_test_class
from .core.exclusions import exclude
from .core.fixtures import parametrize, parameters
from .core.fixtures.parameters import ParametrizationValue as param
from .core.fixtures.utils import fixture, use_fixtures, nofixtures, generator_fixture, yield_fixture, use
from .core.requirements import requires
from .utils import skip_test, skipped, add_error, add_failure, set_test_detail, repeat, register_skip_exception
from .utils.interactive import start_interactive_shell
from .warnings import ignore_warnings, ignored_warnings, clear_ignored_warnings
from .runner import run_tests
import logbook
logger = logbook.Logger(__name__)



from ..utils.python import resolve_underlying_function

_SLASH_REQUIRES_KEY_NAME = '__slash_requirements__'


def requires(req, message=None):
    """A decorator specifying that the decorated tests requires a certain precondition in order to run

    :param req: Either a function receiving no arguments and returning a boolean, or a boolean specifying whether or not
       the requirement is met
    """
    if not isinstance(req, Requirement):
        req = Requirement(req, message)
    else:
        assert message is None, 'Cannot specify message when passing Requirement objects to slash.requires'

    def decorator(func_or_class):
        reqs = _get_requirements_list(func_or_class)
        reqs.append(req)
        return func_or_class
    return decorator

def _get_requirements_list(thing, create=True):

    thing = resolve_underlying_function(thing)
    existing = getattr(thing, _SLASH_REQUIRES_KEY_NAME, None)

    key = id(thing)


    if existing is None or key != existing[0]:
        new_reqs = (key, [] if existing is None else existing[1][:])
        if create:
            setattr(thing, _SLASH_REQUIRES_KEY_NAME, new_reqs)
            assert thing.__slash_requirements__ is new_reqs
        returned = new_reqs[1]
    else:
        returned = existing[1]

    return returned


def get_requirements(test):
    return list(_get_requirements_list(test, create=False))


class Requirement(object):

    def __init__(self, req, message=None):
        super(Requirement, self).__init__()
        self._req = req
        self._message = message

    def __repr__(self):
        if self._message is not None:
            return self._message
        if isinstance(self._req, bool):
            return '?'
        if hasattr(self._req, '__name__'):
            return '<{.__name__}>'.format(self._req)
        return repr(self._req)

    def is_met(self):
        if isinstance(self._req, bool):
            return self._req, self._message
        returned = self._req()
        if not isinstance(returned, tuple):
            returned = (returned, self._message)
        return returned


class Skip(Requirement):
    """
    A special requirement used for implementing @slash.skipped
    """
    def __init__(self, reason=None):
        super(Skip, self).__init__(False, reason)
        self.reason = reason



# pylint: disable=unused-variable,redefined-outer-name
import os
import collections
import warnings
from contextlib import contextmanager
from uuid import uuid4
from vintage import deprecated

import logbook
import pytest
import inspect

import slash

from .utils import run_tests_assert_success, without_pyc


@pytest.mark.parametrize('reprify', [repr, str])
def test_str_repr(warning, reprify):
    assert 'this is a warning' in reprify(warning)


def test_location(warning):
    assert warning.details['filename'] == without_pyc(__file__)


def test_to_dict(warning):
    assert isinstance(warning.to_dict(), dict)


def test_warning_added_hook(suite, suite_test):

    captured = []

    @slash.hooks.register
    def warning_added(warning):
        captured.append(warning)

    suite_test.append_line('slash.logger.warning("message {}", "here")')
    suite.run()
    assert captured
    [w] = captured # pylint: disable=unbalanced-tuple-unpacking
    assert w.message == 'message here'
    assert isinstance(w.lineno, int)
    assert isinstance(w.filename, str)
    assert w.lineno
    assert w.filename
    assert w.filename.rsplit(os.path.sep, 1)[-1] == suite_test.file.get_relative_path()
    warning_type = w.details['type']
    assert isinstance(warning_type, str)
    assert warning_type == 'LogbookWarning'


def _get_current_line_info():
    frame = inspect.currentframe().f_back
    return frame.f_code.co_filename, frame.f_lineno


def test_native_warnings(message):

    def test_example():
        with logbook.TestHandler() as handler:
            filename, line_no = _get_current_line_info()
            warnings.warn(message)
        assert len(handler.records) == 1
        rec_message = handler.records[0].message
        assert rec_message.endswith(message)
        assert filename in rec_message
        assert str(line_no + 1) in rec_message

    s = run_tests_assert_success(test_example)

    assert len(s.warnings) == 1
    [w] = s.warnings
    assert w.message == message


def test_deprecation(message):
    @deprecated(message=message)
    def deprecated_func():
        pass
    with capturing_native_warnings() as handlers:
        deprecated_func()
    assert len(handlers.native_warnings) == 1
    native_warning = handlers.native_warnings[0]
    warning_message = native_warning.message.args[0]
    assert message in warning_message


@pytest.fixture
def message():
    return 'some message here {}'.format(uuid4())


@pytest.fixture
def warning():
    class SampleTest(slash.Test):

        def test(self):
            slash.logger.warning("this is a warning. Param: {0}", 1)

    session = run_tests_assert_success(SampleTest)

    assert len(session.warnings) == 1

    [warning] = session.warnings
    return warning


@contextmanager
def capturing_native_warnings():
    with logbook.TestHandler() as log_handler:
        with warnings.catch_warnings(record=True) as recorded:
            warnings.simplefilter('always')
            handlers = collections.namedtuple('Handlers', ('log', 'native_warnings'))(log_handler, recorded)
            yield handlers

@pytest.mark.parametrize('should_ignore', [True, False])
def test_capture_warnings_before_session_start(request, suite, should_ignore):
    request.addfinalizer(slash.clear_ignored_warnings)
    warning_message = "warning_message"

    @slash.hooks.register
    def configure():    # pylint: disable=unused-variable
        warnings.warn(warning_message)
    if should_ignore:
        slash.ignore_warnings(message=warning_message)
    session_warnings = suite.run().session.warnings
    if should_ignore:
        assert len(session_warnings) == 0  # pylint: disable=len-as-condition
    else:
        assert len(session_warnings) == 1
        assert list(session_warnings)[0].message == warning_message



# pylint: disable=unused-variable, unused-argument
import functools

import pytest
import itertools
import slash
from slash.exceptions import UnknownFixtures

from .utils import make_runnable_tests


def test_resolve_fixture_object():
    with slash.Session() as s:

        @s.fixture_store.add_fixture
        @slash.parametrize('x', [1, 2])
        @slash.fixture
        def fixture1(x):
            pass

        @s.fixture_store.add_fixture
        @slash.parametrize('x', [3, 4])
        @slash.fixture
        def fixture2(fixture1, x):
            pass

        def test_something(fixture2):
            pass

        s.fixture_store.resolve()

        with s.get_started_context():
            tests = make_runnable_tests(test_something)

    _resolve = functools.partial(s.fixture_store.resolve_name, start_point=tests[0])
    def _resolve_values(path):
        return [v.value for v in _resolve(path).values]

    # check simple resolutions
    assert _resolve('fixture2').info is fixture2.__slash_fixture__
    assert _resolve('fixture2.fixture1').info is fixture1.__slash_fixture__

    # check parameter resolution
    assert _resolve_values('fixture2.fixture1.x') == [1, 2]
    assert _resolve_values('fixture2.x') == [3, 4]

    for invalid_name in ['fixture2.x.y']:
        with pytest.raises(UnknownFixtures):
            _resolve(invalid_name)

def test_resolve_fixture_object_namespace_correctness():
    with slash.Session() as s:

        store = s.fixture_store

        @store.add_fixture
        @slash.fixture
        def global_fixture_1(dependency_fixture):
            pass

        @store.add_fixture
        @slash.fixture
        def dependency_fixture():
            pass

        expected = dependency_fixture

        store.push_namespace()

        @store.add_fixture
        @slash.fixture
        def local_fixture(global_fixture_1):
            pass

        @store.add_fixture
        @slash.fixture
        def dependency_fixture():         # pylint: disable=function-redefined
            pass

        def test_something(local_fixture):
            pass

        store.resolve()

        with s.get_started_context():
            tests = make_runnable_tests(test_something)

        test = tests[0]
        resolved = store.resolve_name('local_fixture.global_fixture_1.dependency_fixture', test, namespace=test.get_fixture_namespace()).info
        assert resolved is expected.__slash_fixture__


def test_invalid_name():
    with slash.Session() as s:
        with pytest.raises(UnknownFixtures):
            s.fixture_store.resolve_name('', start_point=object())


def test_resolve_fixture_parameterization_with_scopes(suite_builder):
    # pylint: disable=unused-variable
    @suite_builder.first_file.add_code
    def __code__():
        import slash # pylint: disable=redefined-outer-name, reimported
        @slash.parametrize('x', [1, 2])
        @slash.fixture(scope='session')
        def session_fixture(x):
            return x

        @slash.parametrize('x', [3, 4])
        @slash.fixture(scope='module')
        def module_fixture(x):
            return x

        @slash.parametrize('x', [5, 6])
        @slash.fixture
        def fixture_test(x):
            return x

        def test_1(session_fixture, module_fixture, fixture_test):
            slash.context.result.data['params'] = (session_fixture, module_fixture, fixture_test)

    suite_builder.build().run().assert_success(8).with_data([{'params': x} for x in list(itertools.product(range(1, 3), range(3, 5), range(5, 7)))])


def test_interdependent_fixtures_called_once(suite_builder):
    # See https://github.com/getslash/slash/issues/882
    # pylint: disable=unused-variable, redefined-outer-name,reimported
    @suite_builder.first_file.add_code
    def __code__():
        import slash
        import uuid


        @slash.fixture
        def fixture_2(fixture_1):
            return 'improved-{}'.format(fixture_1)

        @slash.fixture
        @slash.parameters.toggle("toggle")
        def fixture_1(toggle):
            return str(uuid.uuid4())


        def test_1(fixture_1, fixture_2):
            assert fixture_1 in fixture_2

    suite_builder.build().run().assert_success(2)



from __future__ import print_function

import signal
import sys
import traceback
from slash import skip_test, context

from ...utils.debug import launch_debugger
from ..interface import PluginInterface

_PLUGIN_NAME = "signal handling"

_signal_handlers = {}

def register_handler(signal_name, skip_plugin=False):
    def register_handler_func(func):
        signal_no = getattr(signal, signal_name, None)
        if signal_no is not None:
            _signal_handlers[signal_no] = func
            if not skip_plugin:
                from ..plugin_manager import manager
                plugin = manager.get_active_plugins().get(_PLUGIN_NAME)
                if plugin:
                    plugin._set_handler(signal_no, func)  # pylint: disable=protected-access
        return func
    return register_handler_func


@register_handler('SIGUSR1', skip_plugin=True)
def _usr1_handler(*_, **__):
    if not sys.stdout.isatty():
        message = "Running without TTY when {!r} plugin caught SIGUSR1. Current stack:".format(_PLUGIN_NAME)
        if context.session is None:
            print(message)
        else:
            context.session.reporter.report_message(message)
            traceback.print_stack()
    else:
        launch_debugger(exc_info=(None, None, None))


@register_handler('SIGUSR2', skip_plugin=True)
def _usr2_handler(*_, **__):
    skip_test("Skipped due to SIGUSR2 signal caught by {!r} plugin".format(_PLUGIN_NAME))


class Plugin(PluginInterface):

    def __init__(self):
        super(Plugin, self).__init__()
        self._orig_handlers = {}

    def get_name(self):
        return _PLUGIN_NAME

    def _set_handler(self, signal_no, signal_handler):
        # This method is used by register_handler as well
        # It cannot be public due to import-loop while importing "registers_on" before
        # PluginManager initialization finished (it creates all builtin plugins on init)
        self._orig_handlers[signal_no] = signal.signal(signal_no, signal_handler)

    def activate(self):
        super().activate()
        for (sig_no, sig_handler) in _signal_handlers.items():
            self._set_handler(sig_no, sig_handler)

    def deactivate(self):
        super().deactivate()
        for sig_no, sig_handler in self._orig_handlers.items():
            signal.signal(sig_no, sig_handler)
        self._orig_handlers.clear()



from ..conf import config
from .. import hooks
import os
import logbook
from ..utils.tmux_utils import create_new_window, create_new_pane
from slash import ctx
import subprocess
import time
import signal
import errno
import sys

_logger = logbook.Logger(__name__)

def is_process_running(pid):
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            return False
        else:
            raise
    return True

class WorkerConfiguration(object):
    def __init__(self, args, worker_id):
        self._worker_id = worker_id
        self.argv = [sys.executable, '-m', 'slash.frontend.main', args.cmd, '--parallel-parent-session-id', ctx.session.id] + \
                     args.argv + ["--parallel-worker-id", str(worker_id)]
        self._excluded_tests = set()
        self._forced_tests = set()
        self._pid = None

    def get_forced_tests(self):
        return self._forced_tests

    def get_excluded_tests(self):
        return self._excluded_tests

    def get_worker_id(self):
        return self._worker_id

    def _start(self):
        raise NotImplementedError()

    def start(self):
        hooks.before_worker_start(worker_config=self) # pylint: disable=no-member
        _logger.notice("Starting worker number {}", self._worker_id)
        self.argv += ['--parallel-port', str(ctx.session.parallel_manager.server.port),\
                      '--keepalive-port', str(ctx.session.parallel_manager.keepalive_server.port), \
                      '--workers-error-dir', ctx.session.parallel_manager.workers_error_dircetory]
        self._start()

    def kill(self):
        raise NotImplementedError()

    def handle_timeout(self):
        raise NotImplementedError()

    def _test_or_index_to_index(self, test_or_test_index):
        test_index = test_or_test_index.__slash__.parallel_index if hasattr(test_or_test_index, '__slash__') else test_or_test_index
        assert isinstance(test_index, int)
        return test_index

    def exclude_test(self, test_or_test_index):
        """Prevents from the test/test_index to be executed on the specific worker
        """
        test_index = self._test_or_index_to_index(test_or_test_index)
        if test_index in self._forced_tests:
            raise RuntimeError('Cannot exclude test_index {} for client {} since it has been forced before'.format(test_index,
                                                                                                       self._worker_id))
        if test_index in self._excluded_tests:
            _logger.warning('test_index {} already in exclude list for worker {}, not adding it', test_index, self._worker_id)
        else:
            self._excluded_tests.add(test_index)

    def force_test(self, test_or_test_index):
        """Forces the test/test_index to be executed on the specific worker
        """
        test_index = self._test_or_index_to_index(test_or_test_index)
        if test_index in self._excluded_tests:
            raise RuntimeError('Cannot force test_index {} for client {} since it has been excluded before'.format(test_index,
                                                                                                       self._worker_id))
        if test_index in self._forced_tests:
            _logger.warning('test_index {} already in forced list for worker {}, not adding it', test_index, self._worker_id)
        else:
            self._forced_tests.add(test_index)

    def get_pid(self):
        return self._pid

    def is_active(self):
        raise NotImplementedError()

    def wait_to_finish(self):
        raise NotImplementedError()

class ProcessWorkerConfiguration(WorkerConfiguration):
    def _start(self):
        with open(os.devnull, 'w') as devnull:
            self._pid = subprocess.Popen(self.argv, stdin=devnull, stdout=devnull, stderr=devnull)

    def kill(self):
        if self._pid is not None:
            self._pid.send_signal(signal.SIGTERM)

    def handle_timeout(self):
        if self.is_active():
            self._pid.kill()

    def is_active(self):
        return self._pid.poll() is None

    def wait_to_finish(self):
        self._pid.wait()

class TmuxWorkerConfiguration(WorkerConfiguration):
    def __init__(self, basic_args, worker_id):
        super(TmuxWorkerConfiguration, self).__init__(basic_args, worker_id)
        self._window_handle = None

    def _start(self):
        self.argv += [';$SHELL']
        command = ' '.join(self.argv)
        if config.root.tmux.use_panes:
            self._window_handle = create_new_pane(command)
        else:
            self._window_handle = create_new_window("worker {}".format(self._worker_id), command)

    def get_pid(self):
        if self._pid is None:
            self._pid = ctx.session.parallel_manager.server.worker_to_pid.get(self._worker_id)
        return self._pid

    def handle_timeout(self):
        if not config.root.tmux.use_panes:
            self._window_handle.rename_window('stopped_client_{}'.format(self._worker_id))

    def is_active(self):
        return is_process_running(self.get_pid())

    def kill(self):
        try:
            os.kill(self.get_pid(), signal.SIGTERM)
        except OSError as err:
            if err.errno != errno.ESRCH:
                raise

    def wait_to_finish(self):
        for _ in range(10):
            if not self.is_active():
                break
            else:
                time.sleep(0.5)



import munch
from slash.frontend.list_config import list_config
from io import StringIO


def test_slash_list_config():
    report_stream = StringIO()
    list_config(munch.Munch(argv=[]), report_stream)
    assert report_stream.getvalue()

def test_slash_list_config_with_filters():
    report_stream = StringIO()
    list_config(munch.Munch(argv=['log']), report_stream)
    assert 'log.root' in report_stream.getvalue()



# pylint: disable=redefined-outer-name
import pytest

from functools import wraps

from slash.utils.python import call_all_raise_first, resolve_underlying_function


def test_call_all_raise_first(funcs):
    exc_type = funcs[2].raise_when_called()

    with pytest.raises(exc_type):
        call_all_raise_first(funcs)

    for index, func in enumerate(funcs):  # pylint: disable=unused-variable
        assert func.called_count == 1


@pytest.fixture
def funcs():

    class Func(object):

        called_count = 0
        exc_type = None

        def __call__(self):
            self.called_count += 1
            if self.exc_type is not None:
                raise self.exc_type()

        def raise_when_called(self):
            class CustomException(Exception):
                pass

            self.exc_type = CustomException
            return self.exc_type

    return [Func() for _ in range(10)]


@pytest.mark.parametrize('class_method', [True, False])
def test_resolve_underlying_function_method(class_method):
    if class_method:
        decorator = classmethod
    else:
        decorator = lambda f: f

    class Blap(object):

        @decorator
        def method(self):
            pass

    resolved = resolve_underlying_function(Blap.method)
    assert resolved is resolve_underlying_function(Blap.method) # stable
    assert not hasattr(resolved, '__func__')
    assert resolved.__name__ == 'method'


@pytest.mark.parametrize('thing', [object(), object, None, 2, "string"])
def test_resolve_underlying_function_method_no_op(thing):
    assert resolve_underlying_function(thing) is thing


def _example_decorator(func):
    @wraps(func)
    def new_func():
        pass

    return new_func

def test_resolve_underlying_decorator_regular_func():

    def orig():
        pass
    decorated = _example_decorator(orig)
    assert resolve_underlying_function(decorated) is orig

def test_resolve_underlying_decorator_method():

    class Blap(object):

        def orig(self):
            pass

        decorated = _example_decorator(orig)

    assert resolve_underlying_function(Blap.decorated) is resolve_underlying_function(Blap.orig)
    assert resolve_underlying_function(Blap.decorated).__name__ == 'orig'



import os
import time
import logbook
import threading
from tempfile import mkdtemp
from  six.moves import xmlrpc_client
from .. import log
from ..exceptions import INTERRUPTION_EXCEPTIONS, ParallelServerIsDown, ParallelTimeout
from ..conf import config
from .server import Server, ServerStates, KeepaliveServer
from .worker_configuration import TmuxWorkerConfiguration, ProcessWorkerConfiguration

_logger = logbook.Logger(__name__)
log.set_log_color(_logger.name, logbook.NOTICE, 'blue')

TIME_BETWEEN_CHECKS = 2
MAX_CONNECTION_RETRIES = 200

def get_xmlrpc_proxy(address, port):
    return xmlrpc_client.ServerProxy('http://{}:{}'.format(address, port))


class ParallelManager(object):
    def __init__(self, args):
        super(ParallelManager, self).__init__()
        self.server = None
        self.workers_error_dircetory = mkdtemp()
        self.args = args
        self.workers_num = config.root.parallel.num_workers
        self.workers = {}
        self.server_thread = None
        self.keepalive_server = None
        self.keepalive_server_thread = None
        self._create_workers()

    def _create_workers(self):
        for index in range(1, self.workers_num+1):
            _logger.debug("Creating worker number {}", index)
            index_str = str(index)
            worker_cls = TmuxWorkerConfiguration if config.root.tmux.enabled else ProcessWorkerConfiguration
            self.workers[index_str] = worker_cls(self.args, index_str)

    def try_connect(self):
        for _ in range(MAX_CONNECTION_RETRIES):
            if self.server.state != ServerStates.NOT_INITIALIZED and self.keepalive_server.state != ServerStates.NOT_INITIALIZED:
                return
            time.sleep(0.1)
        raise ParallelServerIsDown("Cannot connect to XML_RPC server")

    def start_server_in_thread(self, collected):
        self.server = Server(collected)
        self.server_thread = threading.Thread(target=self.server.serve, args=())
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        self.keepalive_server = KeepaliveServer()
        self.keepalive_server_thread = threading.Thread(target=self.keepalive_server.serve, args=())
        self.keepalive_server_thread.setDaemon(True)
        self.keepalive_server_thread.start()

    def kill_workers(self):
        for worker in list(self.workers.values()):
            worker.kill()

    def report_worker_error_logs(self):
        found_worker_errors_file = False
        for file_name in os.listdir(self.workers_error_dircetory):
            if file_name.startswith(config.root.parallel.worker_error_file):
                found_worker_errors_file = True
                with open(os.path.join(self.workers_error_dircetory, file_name)) as worker_file:
                    content = worker_file.readlines()
                    for line in content:
                        _logger.error("{}: {}", file_name, line, extra={'capture': False})
        if not found_worker_errors_file:
            _logger.error("No worker error files were found", extra={'capture': False})

    def handle_error(self, failure_message):
        _logger.error(failure_message, extra={'capture': False})
        self.kill_workers()
        self.report_worker_error_logs()
        get_xmlrpc_proxy(config.root.parallel.server_addr, self.server.port).report_session_error(failure_message)
        raise ParallelTimeout(failure_message)

    def wait_all_workers_to_connect(self):
        while self.server.state == ServerStates.WAIT_FOR_CLIENTS:
            if time.time() - self.server.start_time > config.root.parallel.worker_connect_timeout * self.workers_num:
                self.handle_error("Timeout: Not all clients connected to server, terminating.\n\
                                   Clients connected: {}".format(self.server.connected_clients))
            time.sleep(TIME_BETWEEN_CHECKS)

    def check_worker_timed_out(self):
        workers_last_connection_time = self.keepalive_server.get_workers_last_connection_time()
        for worker_id in self.server.get_connected_clients():
            worker_last_connection_time = workers_last_connection_time.get(worker_id, None)
            if worker_last_connection_time is None: #worker keepalive thread didn't started yet
                continue
            if time.time() - worker_last_connection_time > config.root.parallel.communication_timeout_secs:
                _logger.error("Worker {} is down, terminating session", worker_id, extra={'capture': False})
                self.report_worker_error_logs()
                self.workers[worker_id].handle_timeout()
                get_xmlrpc_proxy(config.root.parallel.server_addr, self.server.port).report_client_failure(worker_id)

    def check_no_requests_timeout(self):
        if time.time() - self.keepalive_server.last_request_time > config.root.parallel.no_request_timeout:
            _logger.error("No request sent to server for {} seconds, terminating",
                          config.root.parallel.no_request_timeout, extra={'capture': False})
            if self.server.has_connected_clients():
                _logger.error("Clients that are still connected to server: {}",
                              self.server.connected_clients, extra={'capture': False})
            if self.server.has_more_tests():
                _logger.error("Number of unstarted tests: {}", len(self.server.get_unstarted_tests()),
                              extra={'capture': False})
            if self.server.executing_tests:
                _logger.error("Currently executed tests indexes: {}", self.server.executing_tests.values(),
                              extra={'capture': False})
            self.handle_error("No request sent to server for {} seconds, terminating".format(config.root.parallel.no_request_timeout))

    def start(self):
        self.try_connect()
        try:
            for worker in list(self.workers.values()):
                worker.start()
            self.wait_all_workers_to_connect()
            while self.server.should_wait_for_request():
                self.check_worker_timed_out()
                self.check_no_requests_timeout()
                time.sleep(TIME_BETWEEN_CHECKS)
        except INTERRUPTION_EXCEPTIONS:
            _logger.error("Server interrupted, stopping workers and terminating", extra={'capture': False})
            get_xmlrpc_proxy(config.root.parallel.server_addr, self.server.port).session_interrupted()
            self.kill_workers()
            raise
        finally:
            for worker in list(self.workers.values()):
                worker.wait_to_finish()

            get_xmlrpc_proxy(config.root.parallel.server_addr, self.server.port).stop_serve()
            get_xmlrpc_proxy(config.root.parallel.server_addr, self.keepalive_server.port).stop_serve()
            self.server_thread.join()
            self.keepalive_server_thread.join()



from .utils import (
    CustomException,
)
from contextlib import contextmanager

import pytest

import slash
from slash.exceptions import TestFailed as _TestFailed, ExpectedExceptionNotCaught


@pytest.mark.parametrize('pair', [
    (1, 1),
    (1, 1.00000001),
])
def test_assert_almost_equal_positive(pair):
    a, b = pair
    slash.assert_almost_equal(a, b)


@pytest.mark.parametrize('combination', [
    (1, 1, 0),
    (1, 1, 0.1),
    (1, 1.1, 0.5),
    (1.0001, 1.00009, 0.00002),
])
def test_assert_almost_equal_positive_with_delta(combination):
    a, b, delta = combination
    slash.assert_almost_equal(a, b, delta)


@pytest.mark.parametrize('combination', [
    (1, 12, 0),
    (1, 1.1, 0.00001),
    (1.0001, 1.00009, 0.000001),
])
def test_assert_almost_equal_negative_with_delta(combination):
    a, b, delta = combination
    with pytest.raises(AssertionError):
        slash.assert_almost_equal(a, b, delta)


@pytest.mark.parametrize('func', [slash.assert_raises])
def test_assert_raises(func):
    thrown = CustomException()
    with func(CustomException) as caught:
        raise thrown
    assert caught.exception is thrown
    try:
        with func(CustomException):
            raise OtherException()
    except OtherException:
        pass
    else:
        assert False, "func allowed a different type of exception to be raised"
    try:
        with func(CustomException):
            pass
    except ExpectedExceptionNotCaught as e:
        assert " not raised" in str(e)
        assert e.expected_types == (CustomException,)
    else:
        assert False, "assert_raises allowed success"


@pytest.mark.parametrize('func', [slash.assert_raises])
def test_assert_raises_multiple_exceptions(func):
    class CustomException1(Exception):
        pass

    class CustomException2(Exception):
        pass

    class CustomException3(Exception):
        pass

    exception_types = (CustomException1, CustomException2)

    for exc_type in exception_types:
        with func(exception_types) as caught:
            value = exc_type('!')
            raise value
        assert caught.exception is value

    value = CustomException3('!')
    with pytest.raises(CustomException3) as caught:
        with func(exception_types):
            raise value
    assert caught.value is value

# boilerplate


class OtherException(BaseException):
    pass

_MESSAGE = "SOME MESSAGE HERE"


def good(*args):
    _test_assertion(args, positive=True)


def bad(*args):
    _test_assertion(args, positive=False)


def _test_assertion(args, positive):
    positive_assertion = _current_positive_assertion
    negative_assertion = _current_negative_assertion
    if not positive:
        positive_assertion, negative_assertion = negative_assertion, positive_assertion
    positive_assertion(*args)
    positive_assertion(*(args + (_MESSAGE,)))
    for msg in (None, _MESSAGE):
        try:
            negative_assertion(*(args + (msg,)))
        except _TestFailed as e:
            if msg is not None:
                assert msg in str(e)
        else:
            assert False, "Assertion did not fail"


_current_positive_assertion = None
_current_negative_assertion = None


@contextmanager
def checking(assertion, negative_assertion):
    global _current_positive_assertion # pylint: disable=global-statement
    global _current_negative_assertion # pylint: disable=global-statement
    assert _current_positive_assertion is None
    _current_positive_assertion = assertion
    _current_negative_assertion = negative_assertion
    try:
        yield
    finally:
        _current_positive_assertion = None
        _current_negative_assertion = None



# pylint: disable=redefined-outer-name
from uuid import uuid1

import slash
import pytest
from sentinels import NOTHING
from slash.core.local_config import LocalConfig


def test_local_conf(local_conf, expected_dict):
    conf = local_conf.get_dict()
    for key, value in expected_dict.items():
        assert [value] == conf[key]


def test_local_conf_loading_multiple_times(local_conf_dir, expected_dict):
    items = list(expected_dict.items())
    local_conf = LocalConfig()

    local_conf.push_path(str(local_conf_dir.join('..')))
    var_name, var_value = items[0]
    var = local_conf.get_dict()[var_name]
    assert var == [var_value]

    local_conf.push_path(str(local_conf_dir))

    assert var is local_conf.get_dict()[var_name]


def test_local_conf_nonexistent_dir():
    with pytest.raises(RuntimeError):
        LocalConfig().push_path('/nonexistent/dir')


def test_local_conf_with_both_slashconf_file_and_dir(tmpdir):
    with tmpdir.join('slashconf.py').open('w') as f:
        f.write("x = 1")
    conf_dir = tmpdir.join("slashconf")
    conf_dir.mkdir()
    with tmpdir.join('slashconf.py').open('w') as f:
        f.write("y = 2")
    local_config = LocalConfig()
    with pytest.raises(AssertionError) as caught:
        local_config.push_path(str(tmpdir))
    assert str(tmpdir) in str(caught.value)


def test_local_conf_with_slashconf_dir(tmpdir):
    conf_dir = tmpdir.join("slashconf")
    conf_dir.mkdir()
    with conf_dir.join('some_file.py').open('w') as f:
        f.write("x = 42")
    with conf_dir.join("other_file.py").open("w") as f:
        f.write("a = 1\n")
        f.write("b = 2")
    with conf_dir.join("__init__.py").open("w") as f:
        f.write("import slash\n")
        f.write("c = 1")
    local_config = LocalConfig()
    local_config.push_path(str(tmpdir))
    expected = {"a": 1, "b": 2, "c": 1, "x": 42, "slash": slash}
    config_dict = local_config.get_dict()
    got = {key: config_dict.get(key, NOTHING) for key in expected}
    assert expected == got



@pytest.fixture
def local_conf(local_conf_dir):
    returned = LocalConfig()
    returned.push_path(str(local_conf_dir))
    return returned


@pytest.fixture(params=[True, False])
def local_conf_dir(request, tmpdir, expected_dict):
    root_dir = returned = tmpdir
    is_conf_dir = request.param
    if is_conf_dir:
        returned = conf_dir = root_dir.join("slashconf")
        returned.mkdir()
    for i, (key, value) in enumerate(expected_dict.items()):
        returned = returned.join('subdir{}'.format(i))
        returned.mkdir()

        filename = f"{i}.py" if is_conf_dir else "slashconf.py"
        with returned.join(filename).open('w') as f:
            f.write('{} = [{!r}]'.format(key, value))
    if is_conf_dir:
        return conf_dir
    return returned


@pytest.fixture
def expected_dict():
    return dict(('key_{}'.format(i), 'value_{}_{}'.format(i, uuid1())) for i in range(10))



# pylint: disable=redefined-outer-name
import pytest
import slash
from slash.resuming import (CannotResume, get_last_resumeable_session_id, get_tests_from_previous_session)


def address_in_file(test):
    return ("{0.cls.name}.{0.name}" if test.cls is not None else "{.name}").format(test)

def test_resume_no_session():
    with pytest.raises(CannotResume):
        get_tests_from_previous_session("nonexisting_session")

def test_get_last_resumeable_session(suite):
    suite[len(suite) // 2].when_run.fail()
    prev_id = None
    for i in range(5):  # pylint: disable=unused-variable
        results = suite.run()
        assert results.session.id != prev_id
        prev_id = results.session.id
        assert get_last_resumeable_session_id() == results.session.id

def test_resume(suite):
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    for index, test in enumerate(suite):
        if index > fail_index:
            test.expect_not_run()
    result = suite.run(additional_args=['-x'])
    resumed = get_tests_from_previous_session(result.session.id)

    assert len(resumed) + result.session.results.get_num_started() - 1 == len(suite)


def test_resume_with_filter(suite, config_override):
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    for test in suite[fail_index + 1:]:
        test.expect_not_run()
    result = suite.run(additional_args=['-x'])
    resumed = get_tests_from_previous_session(result.session.id)

    filtered_out_test_name = resumed[0].address_in_file
    with slash.Session():
        config_override('run.filter_strings', [f'not {filtered_out_test_name}'])
        tests = slash.loader.Loader().get_runnables(resumed)
    assert filtered_out_test_name not in [test.__slash__.address_in_file for test in tests]


def test_resume_with_parametrization(suite, suite_test):
    num_values1 = 3
    num_values2 = 5
    suite_test.add_parameter(num_values=num_values1)
    suite_test.add_parameter(num_values=num_values2)
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    summary = suite.run()
    resumed = get_tests_from_previous_session(summary.session.id)

    assert len(summary.get_all_results_for_test(suite_test)) == num_values1 * num_values2
    assert len(resumed) == 1
    assert resumed[0].address_in_file == address_in_file(suite[fail_index])

def test_different_folder_no_resume_session_id(suite, tmpdir):  # pylint: disable=unused-argument
    fail_index = len(suite) // 2
    suite[fail_index].when_run.fail()
    suite.run()
    sessoin_id = get_last_resumeable_session_id()

    assert sessoin_id
    with tmpdir.ensure_dir().as_cwd():
        with pytest.raises(CannotResume):
            sessoin_id = get_last_resumeable_session_id()

def test_delete_old_sessions(suite, config_override):
    result = suite.run()
    assert result.session.id == get_last_resumeable_session_id()
    config_override('resume.state_retention_days', 0)
    result = suite.run()
    with pytest.raises(CannotResume):
        get_last_resumeable_session_id()

def test_failed_and_unstarted_first_fails(suite, config_override):
    config_override('resume.failed_first', True)
    config_override('resume.unstarted_first', True)
    result = suite.run()
    with pytest.raises(CannotResume):
        get_tests_from_previous_session(result.session.id)

@pytest.mark.parametrize('failed_first', [True, False])
def test_failed_first_or_unstarted_first(suite, failed_first, config_override):
    fail_index = len(suite) // 2
    skip_index = fail_index - 1
    suite[skip_index].when_run.skip()
    suite[fail_index].when_run.fail()
    for index, test in enumerate(suite):
        if index > fail_index:
            test.expect_not_run()
    result = suite.run(additional_args=['-x'])
    regular_order = get_tests_from_previous_session(result.session.id)
    if failed_first:
        config_override('resume.failed_first', True)
    else:
        config_override('resume.unstarted_first', True)

    order_after_changing_config = get_tests_from_previous_session(result.session.id)
    assert len(order_after_changing_config) + result.session.results.get_num_started() - 2 == len(suite)

    first_test_in_resumed_suite = fail_index if failed_first else skip_index
    assert order_after_changing_config[0].address_in_file == address_in_file(suite[first_test_in_resumed_suite])

    if failed_first:
        assert regular_order[1] == order_after_changing_config[0]
        assert regular_order[0] == order_after_changing_config[1]
        assert regular_order[2:] == order_after_changing_config[2:]
    else:
        failed_test = order_after_changing_config.pop(len(order_after_changing_config)-1)
        order_after_changing_config.insert(1, failed_test)
        assert regular_order == order_after_changing_config

@pytest.mark.parametrize('failed_first', [True, False])
def test_failed_or_unstarted_with_no_such_tests(suite, failed_first, suite_test, config_override):
    if failed_first:
        suite_test.when_run.skip()
    else:
        suite_test.when_run.fail()
    result = suite.run()
    if failed_first:
        config_override('resume.failed_first', True)
    else:
        config_override('resume.unstarted_first', True)
    [resumed_test] = get_tests_from_previous_session(result.session.id)
    assert resumed_test.address_in_file == address_in_file(suite_test)

@pytest.mark.parametrize('failed_only', [True, False])
def test_failed_only_or_unstarted_first(suite, failed_only, config_override):
    fail_index = len(suite) // 2
    skip_index = fail_index - 1
    suite[skip_index].when_run.skip()
    suite[fail_index].when_run.fail()
    for index, test in enumerate(suite):
        if index > fail_index:
            test.expect_not_run()
    result = suite.run(additional_args=['-x'])

    if failed_only:
        config_override('resume.failed_only', True)
        expected_resume_tests_num = 1
        expected_status = 'failed'
    else:
        config_override('resume.unstarted_only', True)
        expected_resume_tests_num = len(suite) - fail_index
        expected_status = 'planned'
    resumed = get_tests_from_previous_session(result.session.id)

    assert len(resumed) == expected_resume_tests_num
    for test in resumed:
        assert test.status == expected_status

def test_resuming_interrupted_session(suite):
    suite[0].when_run.interrupt()
    for index, test in enumerate(suite):
        if index > 0:
            test.expect_deselect()
    results = suite.run(expect_interruption=True)
    resumed = get_tests_from_previous_session(results.session.id)
    assert len(resumed) == len(suite)
    assert set(test.address_in_file for test in resumed) == set(address_in_file(test) for test in suite)



import copy
import logbook
import time
import threading
from enum import Enum
from six.moves import xmlrpc_server
from ..utils.python import unpickle
from .. import log
from ..ctx import context
from ..runner import _get_test_context
from .. import hooks
from ..conf import config
from .tests_distributer import TestsDistributer

_logger = logbook.Logger(__name__)
log.set_log_color(_logger.name, logbook.NOTICE, 'blue')

NO_MORE_TESTS = "NO_MORE_TESTS"
PROTOCOL_ERROR = "PROTOCOL_ERROR"
WAITING_FOR_CLIENTS = "WAITING_FOR_CLIENTS"

class ServerStates(Enum):
    NOT_INITIALIZED = 1
    WAIT_FOR_CLIENTS = 2
    WAIT_FOR_COLLECTION_VALIDATION = 3
    SERVE_TESTS = 4
    STOP_TESTS_SERVING = 5
    STOP_SERVE = 6

class KeepaliveServer(object):
    def __init__(self):
        super(KeepaliveServer, self).__init__()
        self.clients_last_communication_time = {}
        self.last_request_time = time.time()
        self.state = ServerStates.NOT_INITIALIZED
        self.port = None
        self._lock = threading.Lock()

    def keep_alive(self, client_id):
        with self._lock:
            self.clients_last_communication_time[client_id] = self.last_request_time = time.time()
        _logger.debug("Client_id {} sent keep_alive", client_id)

    def stop_serve(self):
        self.state = ServerStates.STOP_SERVE

    def serve(self):
        server = xmlrpc_server.SimpleXMLRPCServer((config.root.parallel.server_addr, 0), allow_none=True, logRequests=False)
        try:
            self.port = server.server_address[1]
            self.state = ServerStates.WAIT_FOR_CLIENTS
            server.register_instance(self)
            _logger.debug("Starting Keepalive server")
            while self.state != ServerStates.STOP_SERVE:
                server.handle_request()
            _logger.debug("Exiting KeepAlive server loop")
        finally:
            server.server_close()

    def get_workers_last_connection_time(self):
        with self._lock:
            return copy.deepcopy(self.clients_last_communication_time)


class Server(object):
    def __init__(self, tests):
        super(Server, self).__init__()
        self.worker_error_reported = False
        self.interrupted = False
        self.state = ServerStates.NOT_INITIALIZED
        self.port = None
        self.tests = tests
        self.worker_session_ids = []
        self.executing_tests = {}
        self.finished_tests = []
        self.num_collections_validated = 0
        self.start_time = time.time()
        self.worker_to_pid = {}
        self.connected_clients = set()
        self.collection = [[test.__slash__.file_path,
                            test.__slash__.function_name,
                            test.__slash__.variation.dump_variation_dict()]
                           for test in self.tests]
        self._sorted_collection = sorted(self.collection)
        self._tests_distrubuter = TestsDistributer(len(self._sorted_collection))

    def has_connected_clients(self):
        return len(self.connected_clients) > 0

    def get_connected_clients(self):
        return self.connected_clients.copy()

    def has_more_tests(self):
        return len(self.finished_tests) < len(self.tests)

    def report_client_failure(self, client_id):
        self.connected_clients.remove(client_id)
        test_index = self.executing_tests.get(client_id, None)
        if test_index is not None:
            _logger.error("Worker {} interrupted while executing test {}", client_id,
                          self.tests[test_index].__slash__.address, extra={'capture': False})
            with _get_test_context(self.tests[test_index], logging=False) as (result, _):
                result.mark_interrupted()
                self.finished_tests.append(test_index)
        self.state = ServerStates.STOP_TESTS_SERVING
        self._mark_unrun_tests()
        self.worker_error_reported = True

    def get_unstarted_tests(self):
        return self._tests_distrubuter.get_unstarted_tests()

    def _mark_unrun_tests(self):
        unstarted_tests_indexes = self._tests_distrubuter.get_unstarted_tests()
        for test_index in unstarted_tests_indexes:
            with _get_test_context(self.tests[test_index], logging=False):
                pass
            self.finished_tests.append(test_index)
        self._tests_distrubuter.clear_unstarted_tests()

    def _get_worker_session_id(self, client_id):
        return "worker_{}".format(client_id)

    def connect(self, client_id, client_pid):
        _logger.notice("Client_id {} connected", client_id)
        self.connected_clients.add(client_id)
        client_session_id = '{}_{}'.format(context.session.id.split('_')[0], client_id)
        context.session.logging.create_worker_symlink(self._get_worker_session_id(client_id), client_session_id)
        hooks.worker_connected(session_id=client_session_id)  # pylint: disable=no-member
        self.worker_session_ids.append(client_session_id)
        self.worker_to_pid[client_id] = client_pid
        self.executing_tests[client_id] = None
        if len(self.connected_clients) >= config.root.parallel.num_workers:
            _logger.notice("All workers connected to server")
            self.state = ServerStates.WAIT_FOR_COLLECTION_VALIDATION

    def validate_collection(self, client_id, sorted_client_collection):
        if not self._sorted_collection == sorted_client_collection:
            _logger.error("Client_id {} sent wrong collection", client_id, extra={'capture': False})
            return False
        self.num_collections_validated += 1
        _logger.debug("Worker {} validated tests successfully", client_id)
        if self.num_collections_validated >= config.root.parallel.num_workers and self.state == ServerStates.WAIT_FOR_COLLECTION_VALIDATION:
            _logger.notice("All workers collected tests successfully, start serving tests")
            self.state = ServerStates.SERVE_TESTS
        return True

    def disconnect(self, client_id, has_failure=False):
        _logger.notice("Client {} sent disconnect", client_id)
        self.connected_clients.remove(client_id)
        if has_failure:
            self.state = ServerStates.STOP_TESTS_SERVING

    def get_test(self, client_id):
        if not self.executing_tests[client_id] is None:
            _logger.error("Client_id {} requested new test without sending former result", client_id,
                          extra={'capture': False})
            return PROTOCOL_ERROR
        if self.state == ServerStates.STOP_TESTS_SERVING:
            return NO_MORE_TESTS
        elif self.state in [ServerStates.WAIT_FOR_CLIENTS, ServerStates.WAIT_FOR_COLLECTION_VALIDATION]:
            return WAITING_FOR_CLIENTS
        elif self.state == ServerStates.SERVE_TESTS and self._tests_distrubuter.has_unstarted_tests():
            test_index = self._tests_distrubuter.get_next_test_for_client(client_id)
            if test_index is None: #we have omre tests but current worker cannot execute them
                return NO_MORE_TESTS
            test = self.tests[test_index]
            self.executing_tests[client_id] = test_index
            hooks.test_distributed(test_logical_id=test.__slash__.id, worker_session_id=self._get_worker_session_id(client_id)) # pylint: disable=no-member
            _logger.notice("#{}: {}, Client_id: {}", test_index + 1, test.__slash__.address, client_id,
                           extra={'highlight': True, 'filter_bypass': True})
            return (self.collection[test_index], test_index)
        else:
            _logger.debug("No unstarted tests, sending end to client_id {}", client_id)
            self.state = ServerStates.STOP_TESTS_SERVING
            return NO_MORE_TESTS

    def finished_test(self, client_id, result_dict):
        _logger.debug("Client_id {} finished_test", client_id)
        test_index = self.executing_tests.get(client_id, None)
        if test_index is not None:
            self.finished_tests.append(test_index)
            self.executing_tests[client_id] = None
            with _get_test_context(self.tests[test_index], logging=False) as (result, _):
                result.deserialize(result_dict)
                context.session.reporter.report_test_end(self.tests[test_index], result)
                if result.has_fatal_exception() or (not result.is_success(allow_skips=True) and config.root.run.stop_on_error):
                    _logger.debug("Server stops serving tests, run.stop_on_error: {}, result.has_fatal_exception: {}",
                                  config.root.run.stop_on_error, result.has_fatal_exception())
                    self.state = ServerStates.STOP_TESTS_SERVING
                    self._mark_unrun_tests()
        else:
            _logger.error("finished_test request from client_id {} with index {}, but no test is mapped to this worker",
                          client_id, test_index, extra={'capture': False})
            return PROTOCOL_ERROR

    def stop_serve(self):
        self.state = ServerStates.STOP_SERVE

    def session_interrupted(self):
        context.session.results.global_result.mark_interrupted()
        self.interrupted = True
        if self.state != ServerStates.STOP_SERVE:
            self.state = ServerStates.STOP_TESTS_SERVING

    def report_warning(self, client_id, pickled_warning):
        _logger.notice("Client_id {} sent warning", client_id)
        try:
            warning = unpickle(pickled_warning)
            context.session.warnings.add(warning)
        except TypeError:
            _logger.error('Error when deserializing warning, not adding it', extra={'capture': False})

    def report_session_error(self, message):
        self.worker_error_reported = True
        _logger.error(message, extra={'capture': False})

    def should_wait_for_request(self):
        return self.has_connected_clients() or self.has_more_tests()

    def serve(self):
        server = xmlrpc_server.SimpleXMLRPCServer((config.root.parallel.server_addr, config.root.parallel.server_port),\
                                                  allow_none=True, logRequests=False)
        try:
            self.port = server.server_address[1]
            self.state = ServerStates.WAIT_FOR_CLIENTS
            server.register_instance(self)
            _logger.debug("Starting server loop")
            while self.state != ServerStates.STOP_SERVE:
                server.handle_request()
            if not self.interrupted:
                context.session.mark_complete()
            _logger.trace('Session finished. is_success={0} has_skips={1}',
                          context.session.results.is_success(allow_skips=True), bool(context.session.results.get_num_skipped()))
            _logger.debug("Exiting server loop")
        finally:
            server.server_close()



def test_basic_use_fixtures(suite, suite_test):
    fixtures_names = []
    events_added = []
    for _ in range(3):
        fixture = suite_test.file.add_fixture()
        events_added.append(fixture.add_event())
        fixtures_names.append(fixture.name)
    suite_test.add_decorator('slash.use_fixtures({})'.format(fixtures_names))
    events_happened = suite.run().events
    for event in events_added:
        assert event in events_happened

def test_nested_fixtures(suite, suite_test):
    nested_fixture = suite_test.file.add_fixture()
    fixture = suite_test.file.add_fixture()
    fixture.depend_on_fixture(nested_fixture)
    suite_test.add_decorator('slash.use_fixtures({})'.format([fixture.name]))
    suite.run()

def test_use_fixture_with_string_raises(suite, suite_test):
    suite_test.add_decorator('slash.use_fixtures("example_str")')
    for test in suite:
        test.expect_deselect()
    suite.run(expect_session_errors=True)

def test_extending(suite, suite_test):
    fixture1, fixture2 = suite_test.file.add_fixture(), suite_test.file.add_fixture()
    event1, event2 = fixture1.add_event(), fixture2.add_event()
    suite_test.add_decorator('slash.use_fixtures(["{}"])'.format(fixture1.name))
    suite_test.add_decorator('slash.use_fixtures(["{}"])'.format(fixture2.name))
    events_happened = suite.run().events
    for event in [event1, event2]:
        assert event in events_happened

def test_parameterize_used_fixutre(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():  # pylint: disable=unused-variable
        import slash

        @slash.fixture
        @slash.parameters.iterate(my_param=['a', 'b'])
        def fixture_1(my_param):  # pylint: disable=unused-variable
            return my_param


        @slash.use_fixtures(["fixture_1"])
        def test_a():  # pylint: disable=unused-variable
            pass

    suite_builder.build().run().assert_results_breakdown(success=2)


def test_fixtures_uses_fixture(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():  # pylint: disable=unused-variable
        import slash

        @slash.fixture
        def fixture_1():
            slash.context.result.data['params'] = ["fixture_1"]

        @slash.fixture
        @slash.use_fixtures(["fixture_3"])
        def fixture_2():
            slash.context.result.data['params'].append("fixture_2")

        @slash.fixture
        def fixture_3():
            slash.context.result.data['params'].append("fixture_3")

        @slash.use_fixtures(["fixture_1", "fixture_2"])
        @slash.fixture
        def fixture_4():
            slash.context.result.data['params'].append("fixture_4")

        @slash.use_fixtures(["fixture_4"])
        def test_a(fixture1: slash.use.fixture_1, fixture_2):  # pylint: disable=unused-variable
            pass

    suite_builder.build().run().assert_success(1).with_data([{'params': ["fixture_1", "fixture_2", "fixture_3",
                                                                         "fixture_4"]}])


def test_fixture_cleanup(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():  # pylint: disable=unused-variable
        import slash # pylint: disable=redefined-outer-name, reimported
        @slash.fixture()
        def fixture_1(): # pylint: disable=unused-variable
            yield
            slash.context.result.data['params'] = 'bla'

        @slash.use_fixtures(["fixture_1"])
        def test_1(): # pylint: disable=unused-variable
            pass

    suite_builder.build().run().assert_success(1).with_data([{'params': 'bla'}])

def test_use_fixture_and_depend_on_it_called_once(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():  # pylint: disable=unused-variable
        import slash # pylint: disable=redefined-outer-name, reimported
        @slash.fixture()
        def fixture_1(): # pylint: disable=unused-variable
            if 'called_count' in slash.context.result.data:
                slash.context.result.data['called_count'] += 1
            else:
                slash.context.result.data['called_count'] = 1

        @slash.use_fixtures(["fixture_1"])
        def test_1(fixture_1): # pylint: disable=unused-variable, unused-argument
            pass

    suite_builder.build().run().assert_success(1).with_data([{'called_count': 1}])



from sentinels import NOTHING

from ..exceptions import TaggingConflict

_TAGS_NAME = "__slash_tags__"


class Tagger(object):
    def __init__(self, tag_name, tag_value):
        self.tag_name = tag_name
        self.tag_value = tag_value

    def __call__(self, target):
        assert callable(target)
        return tag_test(target, self.tag_name, self.tag_value)

    def __rfloordiv__(self, other):
        from .fixtures.parameters import ParametrizationValue

        return ParametrizationValue(value=other, tags=[(self.tag_name, self.tag_value)])


def tag_test(test, tag_name, tag_value):
    tags = get_tags(test)
    if tags is NO_TAGS:
        tags = Tags()
        setattr(test, _TAGS_NAME, tags)
    if tags.get(tag_name, NOTHING) not in (tag_value, NOTHING):
        raise TaggingConflict("Tag {} is already set on {}".format(tag_name, test))
    tags[tag_name] = tag_value
    return test


def tag(tag_name, tag_value=NOTHING):
    """Decorator for tagging tests
    """
    return Tagger(tag_name=tag_name, tag_value=tag_value)


def get_tags(test):
    return getattr(test, _TAGS_NAME, NO_TAGS)


class Tags(object):
    def __init__(self, tags=None):
        super(Tags, self).__init__()
        if tags is None:
            tags = {}
        self._tags = tags

    def __setitem__(self, tag_name, value):
        self._tags[tag_name] = value

    def __getitem__(self, tag_name):
        return self._tags[tag_name]

    def __contains__(self, tag_name):
        return tag_name in self._tags

    has_tag = __contains__

    def _check_conflicting_tags(self, other):
        for (
            tag_name,
            tag_value,
        ) in other._tags.items():  # pylint: disable=protected-access
            if self.get(tag_name, NOTHING) not in (tag_value, NOTHING):
                raise TaggingConflict(
                    "Conflicting tag: {} when adding {}, {}".format(
                        tag_name, self, other
                    )
                )

    def __add__(self, other):
        if other is NO_TAGS:
            return self
        self._check_conflicting_tags(other)
        new_tags = self._tags.copy()
        new_tags.update(other._tags)  # pylint: disable=protected-access
        return Tags(new_tags)

    def copy(self):
        return Tags(self._tags.copy())

    def update(self, other):
        if other is NO_TAGS:
            return
        self._check_conflicting_tags(other)
        self._tags.update(other._tags)  # pylint: disable=protected-access

    def get(self, *args, **kwargs):
        return self._tags.get(*args, **kwargs)

    def matches_pattern(self, pattern, exact=False):
        if "=" in pattern:
            key, predicate = pattern.split("=", 1)
            value = self._tags.get(key, NOTHING)
            return value is not NOTHING and str(value) == predicate

        for key, value in self._tags.items():
            if exact:
                if pattern == key:
                    return True
            else:
                if pattern in key:
                    return True
        return False

    def __iter__(self):
        return iter(self._tags)


class _NoTags(object):
    def __contains__(self, tag_name):
        return False

    def __add__(self, other):
        return other

    has_tag = __contains__

    def matches_pattern(self, pattern, **_):  # pylint: disable=unused-argument
        return False

    def __iter__(self):
        return iter([])


NO_TAGS = _NoTags()



from slash import ctx
from logbook import Logger


_logger = Logger(__name__)

class TestsDistributer(object):
    def __init__(self, num_tests):
        self._unstarted_tests_indices = [i for i in range(num_tests)]
        self._current_tests_index = 0
        workers_config = ctx.session.parallel_manager.workers.values()
        self._workers_excluded_tests = {config.get_worker_id(): config.get_excluded_tests() \
                                        for config in workers_config}
        self._forced_tests_dict = {}
        for config in workers_config:
            for test_index in config.get_forced_tests():
                self._forced_tests_dict[test_index] = config.get_worker_id()
        _logger.debug("forced_tests_dict: {}", self._forced_tests_dict)

    def _can_execute_test(self, test_index, client_id):
        _logger.debug("_can_execute_test - client_id: {}, test_index: {}", client_id, test_index)
        forced_worker = self._forced_tests_dict.get(test_index)
        if forced_worker is not None:
            return forced_worker == client_id
        return test_index not in self._workers_excluded_tests[client_id]

    def get_next_test_for_client(self, client_id):
        ret = None
        for test_index in self._unstarted_tests_indices:
            if self._can_execute_test(test_index, client_id):
                ret = test_index
                break
            else:
                _logger.debug('worker id {} cannot execute test number {}, searching another', client_id, test_index)
        if ret is not None:
            self._unstarted_tests_indices.remove(ret)
        return ret

    def clear_unstarted_tests(self):
        self._unstarted_tests_indices = []

    def get_unstarted_tests(self):
        return list(self._unstarted_tests_indices)

    def has_unstarted_tests(self):
        return len(self._unstarted_tests_indices) > 0



# pylint: disable=redefined-outer-name
import gossip
import pytest
import slash
import slash.core.requirements
from .utils import make_runnable_tests
from .utils.suite_writer.suite import Suite
from .utils.code_formatter import CodeFormatter
from contextlib import ExitStack

from functools import wraps

_UNMET_REQ_DECORATOR = "slash.requires(lambda: False)"
_MET_REQ_DECORATOR = "slash.requires(lambda: True)"


def test_ensure_requirements_called_eagerly(checkpoint1, checkpoint2):
    def predicate_1():
        checkpoint1()
        return False

    def predicate_2():
        checkpoint2()
        return False

    @slash.requires(predicate_1)
    @slash.requires(predicate_2)
    def test_something():
        pass

    with slash.Session() as session:
        with session.get_started_context():
            slash.runner.run_tests(make_runnable_tests(test_something))
    [result] = [
        res for res in session.results.iter_all_results() if not res.is_global_result()
    ]
    assert result.is_skip()
    assert checkpoint1.called
    assert checkpoint2.called


def test_requirements_raises_exception(suite, suite_test):
    @suite_test.file.append_body
    def __code__():  # pylint: disable=unused-variable
        def fail_predicate():  # pylint: disable=unused-variable
            raise Exception("Failing")

    suite_test.add_decorator("slash.requires(fail_predicate)")
    suite_test.expect_error()
    summary = suite.run()
    assert not summary.session.results.is_success(allow_skips=True)


def test_requirements_mismatch_session_success(suite, suite_test):
    suite_test.add_decorator("slash.requires(False)")
    suite_test.expect_skip()
    summary = suite.run()
    assert summary.session.results.is_success(allow_skips=True)


@pytest.mark.parametrize("requirement_fullfilled", [True, False])
@pytest.mark.parametrize("use_message", [True, False])
@pytest.mark.parametrize("use_fixtures", [True, False])
@pytest.mark.parametrize("message_in_retval", [True, False])
def test_requirements(
    suite,
    suite_test,
    requirement_fullfilled,
    use_fixtures,
    use_message,
    message_in_retval,
):

    message = "requires something very important"
    if use_message and message_in_retval:
        retval = "({}, {!r})".format(requirement_fullfilled, message)
    else:
        retval = requirement_fullfilled

    suite_test.add_decorator(
        "slash.requires((lambda: {}), {!r})".format(
            retval, message if use_message and not message_in_retval else ""
        )
    )
    if not requirement_fullfilled:
        suite_test.expect_skip()

    if use_fixtures:
        suite_test.depend_on_fixture(suite.slashconf.add_fixture())
    results = suite.run()
    if requirement_fullfilled:
        assert results[suite_test].is_success()
    else:
        assert not results[suite_test].is_started()
        assert results[suite_test].is_skip()
        if use_message:
            [skip] = results[suite_test].get_skips()
            assert message in skip


def test_requirements_functions_no_message(suite, suite_test):
    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.expect_skip()
    results = suite.run()
    result = results[suite_test]
    [skip] = result.get_skips()
    assert "lambda" in skip


def test_requirements_on_class():
    def req1():
        pass

    def req2():
        pass

    @slash.requires(req1)
    class Test(slash.Test):
        @slash.requires(req2)
        def test_something(self):
            pass

    with slash.Session():
        [test] = make_runnable_tests(Test)  # pylint: disable=unbalanced-tuple-unpacking

    assert set([r._req for r in test.get_requirements()]) == set( # pylint: disable=protected-access
        [req1, req2]
    )


@pytest.fixture
def filename_test_fixture(tmpdir):
    returned = str(tmpdir.join("testfile.py"))

    with open(returned, "w") as f:
        with ExitStack() as stack:
            code = CodeFormatter(f)

            code.writeln("import slash")
            code.writeln("@slash.fixture")
            code.writeln(
                "@slash.requires({}, {})".format(_UNMET_REQ_DECORATOR, '"msg1"')
            )
            code.writeln("def fixture():")
            with code.indented():
                code.writeln("return 1")

            code.writeln("@slash.fixture(autouse=True)")
            code.writeln("@slash.requires({}, {})".format(_MET_REQ_DECORATOR, '"msg2"'))
            code.writeln("def fixture1():")
            with code.indented():
                code.writeln("return 1")

            code.writeln("class Test(slash.Test):")
            stack.enter_context(code.indented())

            code.write("def test_1(")
            code.write("self, ")
            code.writeln("fixture):")
            with code.indented():
                code.writeln("pass")
    return returned


def test_requirements_on_class_with_fixture_and_autouse_fixture(filename_test_fixture):
    with slash.Session():
        [test] = make_runnable_tests( # pylint: disable=unbalanced-tuple-unpacking
            filename_test_fixture
        )
    assert sorted([str(r) for r in test.get_requirements()]) == ["msg1", "msg2"]


def test_unmet_requirements_trigger_avoided_test_hook(suite, suite_test):

    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.expect_skip()

    @gossip.register("slash.test_avoided")
    def test_avoided(reason):  # pylint: disable=unused-variable
        slash.context.result.data["avoided"] = {
            "reason": reason,
            "test_name": slash.context.test.__slash__.address,
        }

    summary = suite.run()
    avoided_result = summary[suite_test]

    for r in summary.session.results.iter_all_results():
        if r is avoided_result:
            assert "avoided" in r.data
            assert "lambda" in r.data["avoided"]["reason"]
            assert "unmet requirement" in r.data["avoided"]["reason"].lower()
            assert r.data["avoided"]["test_name"].split("_")[-1] == suite_test.id
        else:
            assert "avoided" not in r.data


def test_adding_requirement_objects():
    class MyRequirement(slash.core.requirements.Requirement):
        pass

    req = MyRequirement("bla")

    @slash.requires(req)
    def test_something():
        pass

    with slash.Session():
        [test] = make_runnable_tests( # pylint: disable=unbalanced-tuple-unpacking
            test_something
        )  # pylint: disable=unbalanced-tuple-unpacking

    reqs = test.get_requirements()
    assert len(reqs) == 1 and reqs[0] is req


def test_cannot_specify_message_with_requirement_object():
    class MyRequirement(slash.core.requirements.Requirement):
        pass

    with pytest.raises(AssertionError) as caught:
        slash.requires(MyRequirement(""), "message")

    assert "specify message" in str(caught.value)


@pytest.mark.parametrize("is_fixture_requirement_unmet", [True, False])
def test_fixture_and_test_requirements(suite, suite_test, is_fixture_requirement_unmet):
    suite_test.depend_on_fixture(suite.slashconf.add_fixture())
    if is_fixture_requirement_unmet:
        suite_test._fixtures[0][1].add_decorator( # pylint: disable=protected-access
            _UNMET_REQ_DECORATOR
        )
        suite_test.add_decorator(_MET_REQ_DECORATOR)
    else:
        suite_test._fixtures[0][1].add_decorator( # pylint: disable=protected-access
            _MET_REQ_DECORATOR
        )
        suite_test.add_decorator(_UNMET_REQ_DECORATOR)

    suite_test.expect_skip()
    results = suite.run()
    assert results[suite_test].is_skip()
    result = results[suite_test]
    [skip] = result.get_skips()
    assert "lambda" in skip


def test_fixture_of_fixture_requirement(suite, suite_test):
    suite_test.add_decorator(_UNMET_REQ_DECORATOR)
    suite_test.depend_on_fixture(suite.slashconf.add_fixture())
    suite_test._fixtures[0][1].add_decorator( # pylint: disable=protected-access
        _MET_REQ_DECORATOR
    )
    suite_test.expect_skip()
    results = suite.run()
    assert results[suite_test].is_skip()

    result = results[suite_test]
    [skip] = result.get_skips()
    assert "lambda" in skip


def test_autouse_fixture_requirement():
    suite = Suite()
    for _ in range(5):
        test = suite.add_test(type="function")
        test.expect_skip()
    fixture = suite.get_last_file().add_fixture(autouse=True)
    fixture.add_decorator(_UNMET_REQ_DECORATOR)
    suite.run()


def test_class_requirements_siblings(suite_builder):
    @suite_builder.first_file.add_code
    def __code__():  # pylint: disable=unused-variable
        import slash  # pylint: disable=redefined-outer-name, reimported

        @slash.requires(lambda: True)
        class BaseTest(slash.Test):
            pass

        @slash.requires(lambda: False) # pylint: disable=unused-variable
        class FirstTest(BaseTest):
            def test(self):
                pass

        class SecondTest(BaseTest): # pylint: disable=unused-variable
            def test(self):
                pass

    suite_builder.build().run().assert_results_breakdown(skipped=1, success=1)


@pytest.mark.parametrize("class_can_run", [True, False])
@pytest.mark.parametrize("method_can_run", [True, False])
def test_class_requirements_class_and_method(class_can_run, method_can_run):
    @slash.requires(lambda: class_can_run)
    class Test(slash.Test):
        @slash.requires(lambda: method_can_run)
        def test(self):
            pass

    with slash.Session() as session:
        with session.get_started_context():
            slash.runner.run_tests(make_runnable_tests(Test))
    results = [res for res in session.results.iter_test_results()]
    assert len(results) == 1
    if class_can_run and method_can_run:
        assert results[0].is_success()
    else:
        assert results[0].is_skip()


def test_attach_requirements_through_wraps_on_function():

    def decorator(func):
        @wraps(func)
        def new_func():
            pass
        return new_func

    req1, req2, req3, req4 = requirements = [object() for _ in range(4)]

    @slash.requires(req4)
    @slash.parametrize('x', [1, 2, 3])
    @slash.requires(req3)
    @slash.parametrize('y', [2, 3, 4])
    @slash.requires(req2)
    @decorator
    @slash.requires(req1)
    def test_something(x, y): # pylint: disable=unused-argument
        pass

    found_reqs = slash.core.requirements.get_requirements(test_something)
    assert len(found_reqs) == len(requirements)
    for expected, found in zip(requirements, found_reqs):
        assert expected is found._req # pylint: disable=trailing-whitespace, protected-access



import functools
import re
import os
import pytest
import warnings

import slash


def _warn(*args, **kwargs):
    return functools.partial(warnings.warn, *args, **kwargs)


def _catch(*args, **kwargs):
    return functools.partial(slash.ignore_warnings, *args, **kwargs)


class CustomWarning(UserWarning):
    pass


@pytest.mark.parametrize('emitter,catch,can_test_negative', [
    (_warn('hello'), _catch(message='hello'), True),
    (_warn('hello'), _catch(message=re.compile('^hello$')), True),
    (_warn('message', category=CustomWarning), _catch(category=CustomWarning), True),
    (_warn('message'), _catch(filename=__file__), False),
    (_warn('message'), _catch(filename=re.compile('^{}$'.format(re.escape(__file__)))), False),
])
def test_ignore_warnings(emitter, catch, can_test_negative):
    catch()

    with slash.Session() as session:
        emitter()
    assert list(session.warnings) == []

    if can_test_negative:
        with slash.Session() as session:
            warnings.warn('uncaught', category=DeprecationWarning)
        caught = list(session.warnings)
        assert len(caught) == 1
        assert caught[0].message == 'uncaught'


_WARN_MESSAGE = 'Hello World'
@pytest.mark.parametrize('catch,should_ignore', [
    (_catch(message=_WARN_MESSAGE, filename=__file__), True),
    (_catch(message=re.compile('^Hello.*$'), filename=__file__), True),
    (_catch(message=_WARN_MESSAGE, filename=re.compile('^{}'.format(re.escape(os.path.dirname(__file__))))), True),
    (_catch(message=re.compile('^Hello.*$'), filename=re.compile('^{}'.format(re.escape(os.path.dirname(__file__))))), True),
    # Negative  (OR relation instead of AND)
    (_catch(message=_WARN_MESSAGE, category=DeprecationWarning), False),
    (_catch(message=re.compile('^Hello.*$'), category=DeprecationWarning), False),
    (_catch(filename=__file__, category=DeprecationWarning), False),
    (_catch(filename=re.compile('^{}$'.format(re.escape(__file__))), category=DeprecationWarning), False),
])
def test_ignore_warnings_with_multiple_criteria(catch, should_ignore):
    catch()

    with slash.Session() as session:
        warnings.warn(category=CustomWarning, message=_WARN_MESSAGE)
    if should_ignore:
        assert not session.warnings
    else:
        assert len(session.warnings) == 1
        [caught] = list(session.warnings)
        assert caught.message == _WARN_MESSAGE
        assert caught.category == CustomWarning


def test_ignore_warnigns_with_no_parameter():
    slash.ignore_warnings()
    with slash.Session() as session:
        warnings.warn(category=CustomWarning, message=_WARN_MESSAGE)
    assert not session.warnings


@pytest.mark.parametrize('should_clear_warnings', [True, False])
def test_ignored_warnings(should_clear_warnings):
    message = 'my warning'
    with slash.Session() as session:
        with slash.ignored_warnings(message=message):
            warnings.warn(message=message)
            assert not session.warnings
            if should_clear_warnings:
                slash.clear_ignored_warnings()

        warnings.warn(message=message)
        assert len(session.warnings) == 1
        [caught] = list(session.warnings)
        assert caught.message == message


@pytest.fixture(autouse=True)
def warnings_cleanup(request):
    request.addfinalizer(slash.clear_ignored_warnings)



import sys
import slash

from slash.plugins import manager as plugin_manager


def test_notification_formatting():
    messages = []

    @slash.hooks.prepare_notification.register # pylint: disable=no-member
    def check_notification(message): # pylint: disable=unused-variable
        messages.append(message)

    plugin = plugin_manager.get_plugin('notifications')

    with slash.Session():
        try:
            raise Exception('Some message containing {curly_braces}')
        except Exception: # pylint: disable=broad-except
            plugin.entering_debugger(sys.exc_info())
    assert len(messages) == 1
    assert '{curly_braces}' in messages[0].get_short_message()



import os
import logbook
from sqlalchemy import Column, DateTime, String, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import OperationalError
from contextlib import contextmanager, closing
from datetime import datetime, timedelta
from slash.plugins import manager
from ast import literal_eval
from .utils.path import ensure_directory
from .conf import config
from .__version__ import __backslash_version__

_DB_NAME = 'resume_state_v2.db'
_logger = logbook.Logger(__name__)
Base = declarative_base()
_session = sessionmaker()
_db_initialized = False

class ResumeState(Base):
    __tablename__ = 'resume_state'
    id = Column(Integer, primary_key=True)
    session_id = Column(String, nullable=False, index=True)
    file_name = Column(String, nullable=False, index=True)
    address_in_file = Column(String, nullable=False, index=True)
    variation = Column(String, index=True)
    status = Column(String, nullable=False, index=True)


class SessionMetadata(Base):
    __tablename__ = 'session_metadata'
    session_id = Column(String, primary_key=True)
    src_folder = Column(String, nullable=False, index=True)
    created_at = Column(DateTime, nullable=False, index=True)


class ResumeTestStatus(object):
    PLANNED = 'planned'
    SUCCESS = 'success'
    FAILED = 'failed'


class ResumedTestData(object):

    def __init__(self, file_name, address_in_file, variation=None, status=None):
        self.file_name = file_name
        self.address_in_file = address_in_file
        self.variation = variation
        self.status = status

    def __repr__(self):
        return '<ResumedTestData({!r}, {!r}, {!r}, {!r})>'.format(self.file_name, self.address_in_file, self.variation, self.status)

    def __eq__(self, other):
        return self.file_name == other.file_name and self.address_in_file == other.address_in_file and \
               self.variation == other.variation and self.status == other.status

def _init_db():
    resume_dir = os.path.expanduser(config.root.run.resume_state_path)
    ensure_directory(resume_dir)
    engine = create_engine('sqlite:///{}/{}'.format(resume_dir, _DB_NAME))
    _session.configure(bind=engine)
    Base.metadata.create_all(engine)


@contextmanager
def connecting_to_db():
    global _db_initialized  # pylint: disable=global-statement
    if not _db_initialized:
        _init_db()
        _db_initialized = True
    with closing(_session()) as new_session:
        yield new_session
        max_retries = 3
        for i in range(max_retries):
            try:
                new_session.commit() # pylint: disable=no-member
            except OperationalError:
                if i == max_retries - 1:
                    raise


def clean_old_entries():
    # pylint: disable=no-member
    with connecting_to_db() as conn:
        old_sessions_query = conn.query(SessionMetadata).filter \
            (SessionMetadata.created_at < datetime.now() - timedelta(days=config.root.resume.state_retention_days))
        old_sessions_ids = [session.session_id for session in old_sessions_query.all()]
        if old_sessions_ids:
            conn.query(ResumeState).filter(ResumeState.session_id.in_(old_sessions_ids)).delete(synchronize_session=False)
        old_sessions_query.delete(synchronize_session=False)


def save_resume_state(session_result, collected_tests):
    session_metadata = SessionMetadata(
        session_id=session_result.session.id,
        src_folder=os.path.abspath(os.getcwd()),
        created_at=datetime.now())

    tests_to_resume = []
    for result in session_result.iter_test_results():
        metadata = result.test_metadata
        test_to_resume = ResumeState(session_id=session_result.session.id, file_name=metadata.file_path, address_in_file=metadata.address_in_file)
        test_to_resume.variation = str(metadata.variation.id) if metadata.variation else None
        if result.is_success_finished():
            test_to_resume.status = ResumeTestStatus.SUCCESS
        elif not result.is_started() or result.is_skip():
            test_to_resume.status = ResumeTestStatus.PLANNED
        else:
            test_to_resume.status = ResumeTestStatus.FAILED
        tests_to_resume.append(test_to_resume)
    tests_with_no_results = {test.__slash__ for test in collected_tests if session_result.safe_get_result(test) is None}

    for test_metadata in tests_with_no_results:
        test_to_resume = ResumeState(session_id=session_result.session.id, file_name=test_metadata.file_path,\
                                     address_in_file=test_metadata.address_in_file, status=ResumeTestStatus.PLANNED)
        test_to_resume.variation = str(test_metadata.variation.id) if test_metadata.variation else None
        tests_to_resume.append(test_to_resume)

    with connecting_to_db() as conn:
        conn.add(session_metadata) # pylint: disable=no-member
        conn.add_all(tests_to_resume) # pylint: disable=no-member
    _logger.debug('Saved resume state to DB')


def get_last_resumeable_session_id():
    current_folder = os.path.abspath(os.getcwd())
    with connecting_to_db() as conn:
         # pylint: disable=no-member
        session_id = conn.query(SessionMetadata).filter(SessionMetadata.src_folder == current_folder) \
                                                .order_by(SessionMetadata.created_at.desc()).first()
        if not session_id:
            raise CannotResume("No sessions found for folder {}".format(current_folder))
        return session_id.session_id

def _get_resume_status_from_backslash_status(backslash_status):
    if backslash_status in ['FAILURE', 'ERROR', 'INTERRUPTED']:
        return ResumeTestStatus.FAILED
    elif backslash_status in ['SUCCESS']:
        return ResumeTestStatus.SUCCESS
    else:
        return ResumeTestStatus.PLANNED

def get_tests_from_remote(session_id, get_successful_tests):
    backslash_plugin = manager.get_active_plugins().get('backslash', None)
    if not backslash_plugin:
        raise CannotResume("Could not find backslash plugin")
    if not hasattr(backslash_plugin, 'is_session_exist') or not hasattr(backslash_plugin, 'get_tests_to_resume'):
        raise CannotResume("Backslash plugin doesn't support remote resuming")
    if not backslash_plugin.is_session_exist(session_id):
        raise CannotResume("Could not find resume data for session {}".format(session_id))
    backslash_version = tuple([int(x) for x in __backslash_version__.split(".")[:2]])
    kwargs = {'filters_dict': {'show_successful': get_successful_tests}} if backslash_version >= (2, 38) else {'get_successful': get_successful_tests}
    remote_tests = backslash_plugin.get_tests_to_resume(session_id, **kwargs)
    return [ResumedTestData(test.info['file_name'],
                            test.info['name'],
                            variation=test.variation,
                            status=_get_resume_status_from_backslash_status(test.status))
            for test in remote_tests]

def get_tests_locally(session_id):
    resumed_tests = []
    with connecting_to_db() as conn:
        # pylint: disable=no-member
        if conn.query(SessionMetadata).filter(SessionMetadata.session_id == session_id).first():
            local_tests = conn.query(ResumeState).filter(ResumeState.session_id == session_id).all()
            resumed_tests = [ResumedTestData(test.file_name,
                                             test.address_in_file,
                                             variation=literal_eval(test.variation) if test.variation else None,
                                             status=test.status)
                            for test in local_tests]
    return resumed_tests

def _filter_tests(tests, get_successful_tests):
    unstarted_only, failed_only = config.root.resume.unstarted_only, config.root.resume.failed_only
    if unstarted_only and failed_only:
        raise CannotResume("Got both unstarted_only and failed_only, cannot resume")
    filtered_status = [ResumeTestStatus.SUCCESS] if get_successful_tests else []
    if unstarted_only:
        filtered_status.append(ResumeTestStatus.PLANNED)
    elif failed_only:
        filtered_status.append(ResumeTestStatus.FAILED)
    else:
        filtered_status.extend([ResumeTestStatus.FAILED, ResumeTestStatus.PLANNED])
    return list(filter(lambda test: test.status in filtered_status, tests))

def _sort_tests(tests):
    unstarted_first, failed_first = config.root.resume.unstarted_first, config.root.resume.failed_first
    if unstarted_first and failed_first:
        raise CannotResume("Got both failed_first and planned_first, cannot choose")
    if failed_first or unstarted_first:
        wanted_status_first = ResumeTestStatus.PLANNED if unstarted_first else ResumeTestStatus.FAILED
        first_tests = list(filter(lambda test: test.status in wanted_status_first, tests))
        other_tests = list(filter(lambda test: test.status not in wanted_status_first, tests))
        return first_tests + other_tests
    return tests

def get_tests_from_previous_session(session_id, get_successful_tests=False):
    resumed_tests = get_tests_locally(session_id)
    if not resumed_tests:
        _logger.debug('No local entry for session {}, searching remote session', session_id)
        resumed_tests = get_tests_from_remote(session_id, get_successful_tests)

    return _sort_tests(_filter_tests(resumed_tests, get_successful_tests))
class CannotResume(Exception):
    pass



import functools
import itertools
import os
import pickle
import sys
from numbers import Number
import gossip
import logbook
from datetime import datetime, timedelta
from vintage import deprecated

from .. import hooks
from ..ctx import context
from ..exception_handling import capture_sentry_exception, handling_exceptions
from .. import exceptions
from ..utils.exception_mark import ExceptionMarker
from ..utils.interactive import notify_if_slow_context
from ..utils.python import unpickle
from .details import Details
from .error import Error

_logger = logbook.Logger(__name__)

_ADDED_TO_RESULT = ExceptionMarker('added_to_result')


class Result(object):
    """Represents a single result for a test which was run
    """
    pickle_key_blacklist = {'test_metadata', 'facts'}

    def __init__(self, test_metadata=None):
        super(Result, self).__init__()
        self.test_metadata = test_metadata
        #: dictionary to be use by tests and plugins to store result-related information for later analysis
        self.data = {}
        self._start_time = None
        self._end_time = None
        self._errors = []
        self._failures = []
        self._skips = []
        #: a :class:`slash.core.details.Details` instance for storing additional test details
        self.details = Details()
        self.facts = Details(set_callback=self._fact_set_callback)
        self._started = False
        self._finished = False
        self._interrupted = False
        self._log_path = None
        self._extra_logs = []

    def _fact_set_callback(self, fact_name, fact_value):
        gossip.trigger('slash.fact_set', name=fact_name, value=fact_value)

    def is_global_result(self):
        return False

    def serialize(self):
        serialized_object = {}
        for key, value in vars(self).items():
            if key not in Result.pickle_key_blacklist:
                try:
                    serialized_object[key] = pickle.dumps(value)
                except (pickle.PicklingError, TypeError):
                    _logger.debug('Failed serializing result, skipping this value. key = {}', key)
                    capture_sentry_exception()
        return serialized_object

    def deserialize(self, result_dict):
        for key, value in result_dict.items():
            try:
                self.__dict__[key] = unpickle(value)
            except TypeError:
                _logger.error('Error when deserialize reult, skipping this value. key = {}', key,
                              extra={'capture': False})
        for failure in self._failures:
            self.add_failure(failure, append=False)
        for error in self._errors:
            self.add_error(error, append=False)
        for skip in self._skips:
            self.add_skip(skip, append=False)

    def add_exception(self, exc_info=None):
        """Adds the currently active exception, assuming it wasn't already added to a result
        """
        if exc_info is None:
            exc_info = sys.exc_info()
        _, exc_value, _ = exc_info  # pylint: disable=unpacking-non-sequence

        if _ADDED_TO_RESULT.is_exception_marked(exc_value):
            return

        _ADDED_TO_RESULT.mark_exception(exc_value)
        if isinstance(exc_value, exceptions.FAILURE_EXCEPTION_TYPES):
            self.add_failure()
        elif isinstance(exc_value, context.session.get_skip_exception_types()):
            self.add_skip(getattr(exc_value, 'reason', str(exc_value)))
        elif isinstance(exc_value, exceptions.INTERRUPTION_EXCEPTIONS):
            err = Error.capture_exception(exc_info=exc_info)
            hooks.interruption_added(result=self, exception=err) # pylint: disable=no-member
            session_result = context.session.results.global_result
            interrupted_test = self.is_interrupted()
            interrupted_session = session_result.is_interrupted()
            if not self.is_global_result():
                self.mark_interrupted()
                if not interrupted_test and not context.session.has_children():
                    with notify_if_slow_context(message="Cleaning up test due to interrupt. Please wait..."),\
                         handling_exceptions(swallow=True):

                        hooks.test_interrupt() # pylint: disable=no-member
            if not interrupted_session:
                session_result.mark_interrupted()

        elif not isinstance(exc_value, GeneratorExit):
            #skip keyboardinterrupt and system exit
            self.add_error(exc_info=exc_info)
        else:
            _logger.trace('Ignoring GeneratorExit exception')

    def has_errors_or_failures(self):
        return bool(self._failures or self._errors)

    def get_log_path(self):
        """Returns log path
        """
        return self._log_path

    def set_log_path(self, path):
        """Set log path
        """
        self._log_path = path

    def get_log_dir(self):
        """Returns log's directory.
        """
        if self._log_path is None:
            return None
        return os.path.dirname(self._log_path)

    def add_extra_log_path(self, path):
        """Add additional log path. This path will be added to the list returns by get_log_paths
        """
        self._extra_logs.append(path)

    def get_log_paths(self):
        """Returns a list of all log paths
        """
        logs = []
        if self._log_path:
            logs.append(self._log_path)
        return logs + list(self._extra_logs)

    def is_started(self):
        return self._started

    def is_not_run(self):
        return not self.is_started() and not self.has_errors_or_failures()

    def mark_started(self):
        self._start_time = datetime.now()
        self._started = True

    def is_error(self):
        return bool(self._errors)

    @property
    def test_id(self):
        return self.test_metadata.id

    def is_failure(self):
        return bool(self._failures)

    def is_just_failure(self):
        """Indicates this is a pure failure, without errors involved"""
        return self.is_failure() and not self.is_error()

    def is_skip(self):
        return bool(self._skips)

    def is_run_and_skip(self):
        return self.is_started() and self.is_skip()

    def is_success(self, allow_skips=False):
        if self._errors or self._failures or self._interrupted:
            return False
        if self._skips:
            return allow_skips
        return self.is_started()

    def is_success_finished(self):
        return self.is_success() and self.is_finished()

    def is_finished(self):
        return self._finished

    def mark_finished(self):
        self._end_time = datetime.now()
        self._finished = True

    def mark_interrupted(self):
        self._interrupted = True

    def is_interrupted(self):
        return self._interrupted

    def add_error(self, e=None, frame_correction=0, exc_info=None, append=True):
        """Adds a failure to the result
        """
        err = self._add_error(self._errors, e, frame_correction=frame_correction + 1, exc_info=exc_info, append=append)
        context.reporter.report_test_error_added(context.test, err)
        return err

    def add_failure(self, e=None, frame_correction=0, exc_info=None, append=True):
        """Adds a failure to the result
        """
        err = self._add_error(self._failures, e, frame_correction=frame_correction + 1, exc_info=exc_info, is_failure=True, append=append)
        context.reporter.report_test_failure_added(context.test, err)
        return err

    def set_test_detail(self, key, value):
        """Adds a generic detail to this test result, which can be later inspected or used
        """
        self.details.set(key, value)

    def _add_error(self, error_list, error=None, frame_correction=0, exc_info=None, is_failure=False, append=True):
        try:
            if error is None:
                error = Error.capture_exception(exc_info=exc_info)
                if error is None:
                    raise RuntimeError('add_error() must be called with either an argument or in an active exception')
            if not isinstance(error, Error):
                error = Error(error, frame_correction=frame_correction + 1, exc_info=exc_info)
            if is_failure:
                # force the error object to be marked as failure
                error.mark_as_failure()
            error.log_added()
            if append:
                error_list.append(error)
            #report children errors or parent session errors only
            if not context.session or not context.session.has_children() or self.is_global_result():
                hooks.error_added(result=self, error=error)  # pylint: disable=no-member
            error.forget_exc_info()
            return error
        except Exception:
            _logger.error("Failed to add error to result", exc_info=True, extra={'capture': False})
            raise

    def add_skip(self, reason, append=True):
        if append:
            self._skips.append(reason)
        context.reporter.report_test_skip_added(context.test, reason)

    def get_duration(self):
        """Returns the test duration time as timedelta object

        :return: timedelta
        """
        if self._end_time is None or self._start_time is None:
            return timedelta()
        return self._end_time - self._start_time

    def get_errors(self):
        """Returns the list of errors recorded for this result

        :return: a list of :class:`slash.core.error.Error` objects
        """
        return self._errors

    def get_failures(self):
        """Returns the list of failures recorded for this result

        :return: a list of :class:`slash.core.error.Error` objects
        """
        return self._failures

    @deprecated('Use result.details.all()', since='0.20.0')
    def get_additional_details(self):
        return self.details.all()

    def get_skips(self):
        return self._skips

    def has_skips(self):
        return bool(self._skips)

    def has_fatal_errors(self):
        return any(e.is_fatal() for e in
                   itertools.chain(self._errors, self._failures))

    has_fatal_exception = has_fatal_errors

    def __repr__(self):
        return "<{} ({})>".format(
            self.__class__.__name__,
            ", ".join(
                attr
                for attr in ("success", "error", "failure", "skip", "finished", "interrupted")
                if getattr(self, "is_{}".format(attr))()
            )
        )


class GlobalResult(Result):

    def __init__(self, session_results=None):
        super(GlobalResult, self).__init__()
        self._session_results = session_results

    def is_global_result(self):
        return True

    def is_success(self, allow_skips=False):
        if not super(GlobalResult, self).is_success(allow_skips=allow_skips):
            return False
        if self._session_results is None:
            return True
        if self._session_results.session.has_children() and self._session_results.session.parallel_manager.server.worker_error_reported:
            return False
        return all(result.is_success(allow_skips=allow_skips) for result in self._session_results.iter_test_results())


class SessionResults(object):

    def __init__(self, session):
        super(SessionResults, self).__init__()
        self.session = session
        self.global_result = GlobalResult(self)
        self._results_dict = {}
        self._iterator = functools.partial(iter, self._results_dict.values())

    def __len__(self):
        return len(self._results_dict)

    def __repr__(self):
        return '<Results: {} successful, {} errors, {} failures, {} skips>'.format(
            self.get_num_successful(),
            self.get_num_errors(),
            self.get_num_failures(),
            self.get_num_skipped())

    def iter_all_additional_details(self):
        for result in self.iter_all_results():
            if result.details:
                yield result, result.get_additional_details()

    def iter_all_failures(self):
        """Iterates over all results which have failures

        yields tuples of the form (result, failure_list)
        """
        for result in self.iter_all_results():
            if result.get_failures():
                yield result, result.get_failures()

    def iter_all_errors(self):
        """Iterates over all results which have errors

        yields tuples of the form (result, errors_list)
        """
        for result in self.iter_all_results():
            if result.get_errors():
                yield result, result.get_errors()

    @property
    def current(self):
        """Obtains the currently running result, if exists

        Otherwise, returns the global result object
        """
        return context.result

    def __iter__(self):
        return self._iterator()

    def is_success(self, allow_skips=False):
        """Indicates whether this run is successful

        :param allow_skips: Whether to consider skips as unsuccessful
        """
        return self.global_result.is_success(allow_skips=allow_skips)

    def is_interrupted(self):
        """Indicates if this session was interrupted
        """
        return any(result.is_interrupted() for result in self._iterator())

    def get_num_results(self):
        return len(self._results_dict)

    def get_num_started(self):
        return self._count(Result.is_started, include_global=False)

    def get_num_successful(self):
        return self._count(Result.is_success_finished, include_global=False)

    def get_num_errors(self):
        return self._count(Result.is_error)

    def get_num_failures(self):
        return self._count(Result.is_just_failure)

    def get_num_skipped(self, include_not_run=True):
        if include_not_run:
            return self._count(Result.is_skip)
        return self._count(Result.is_run_and_skip)

    def get_num_not_run(self):
        return self._count(Result.is_not_run, include_global=False)

    def has_fatal_errors(self):
        """Indicates whether any result has an error marked as fatal (causing the session to terminate)
        """
        return bool(self._count(Result.has_fatal_errors))

    def _count(self, pred, include_global=True):
        returned = 0
        iterator = self.iter_all_results(
        ) if include_global else self.iter_test_results()
        for result in iterator:
            if pred(result):
                returned += 1
        return returned

    def iter_test_results(self):
        """Iterates over results belonging to tests
        """
        return iter(self)

    def iter_all_results(self):
        """Iterates over all results, ending with the global result object
        """
        return itertools.chain(self.iter_test_results(), [self.global_result])

    def create_result(self, test):
        if test.__slash__.id in self._results_dict:
            raise exceptions.SlashInternalError("{} shouldn't appear in results dict before adding its result".format(test.__slash__.id))

        returned = Result(test.__slash__)
        self._results_dict[test.__slash__.id] = returned
        return returned

    def get_result(self, test):
        """Returns the result stored belonging to a given test
        """
        if test.__slash__ is None:
            raise LookupError("Could not find result for {}".format(test))
        return self._results_dict[test.__slash__.id]

    def safe_get_result(self, test):
        if test.__slash__ is None:
            return None
        return self._results_dict.get(test.__slash__.id)

    def __getitem__(self, test):
        if isinstance(test, Number):
            try:
                return next(itertools.islice(self._results_dict.values(), test, test + 1))
            except StopIteration:
                raise IndexError()
        return self.get_result(test)
